# @auto-fold regex /^\s*if/ /^\s*else/ /^\s*def/
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# my coding convention
# **EVAL : evaluate the performance of this method
# **RED  : redo this
# **DEB  : debugging needed in this part
# **DEL  : DELETE AT SOME POINT

import numpy as np
import os
try:
    import h5py
except ImportError:
    h5py = None

from .state import State, PTState

from .backend import Backend_plus

class HDFBackend_plus(Backend_plus):
    """A backend that stores the chain in an HDF5 file using h5py"""
    
    def __init__(
        self,
        filename,
        name="mcmc",
        read_only=False,
        dtype=None,
        compression=None,
        compression_opts=None,
    ):
        if h5py is None:
            raise ImportError("you must install 'h5py' to use the HDFBackend")
        self.filename = filename
        self.name = name
        self.read_only = read_only
        self.compression = compression
        self.compression_opts = compression_opts
        if dtype is None:
            self.dtype_set = False
            self.dtype = np.float64
        else:
            self.dtype_set = True
            self.dtype = dtype

    @property
    def initialized(self):
        if not os.path.exists(self.filename):
            return False
        try:
            with self.open() as f:
                return self.name in f
        except (OSError, IOError):
            return False


    def open(self, mode="r"):
        if self.read_only and mode != "r":
            raise RuntimeError(
                "The backend has been loaded in read-only "
                "mode. Set `read_only = False` to make "
                "changes."
            )
        f = h5py.File(self.filename, mode)
        if not self.dtype_set and self.name in f:
            g = f[self.name]
            if "chain" in g:
                self.dtype = g["chain"].dtype
                self.dtype_set = True
        return f
    

    def reset(self, nwalkers, ndim):
        """Clear the state of the chain and empty the backend

        Args:
            nwakers (int): The size of the ensemble
            ndim (int): The number of dimensions

        """
        with self.open("a") as f:
            if self.name in f:
                del f[self.name]

            g = f.create_group(self.name)
            g.attrs["nwalkers"] = nwalkers
            g.attrs["ndim"] = ndim
            g.attrs["has_blobs"] = False
            g.attrs["iteration"] = 0
            g.create_dataset(
                "accepted",
                data=np.zeros(nwalkers),
                compression=self.compression,
                compression_opts=self.compression_opts,
            )
            g.create_dataset(
                "chain",
                (0, nwalkers, ndim),
                maxshape=(None, nwalkers, ndim),
                dtype=self.dtype,
                compression=self.compression,
                compression_opts=self.compression_opts,
            )
            g.create_dataset(
                "log_prob",
                (0, nwalkers),
                maxshape=(None, nwalkers),
                dtype=self.dtype,
                compression=self.compression,
                compression_opts=self.compression_opts,
            )
            g.create_dataset(
                "log_like",
                (0, nwalkers),
                maxshape=(None, nwalkers),
                dtype=self.dtype,
                compression=self.compression,
                compression_opts=self.compression_opts,
            )
            g.create_dataset(
                "beta_history",
                (0,),
                maxshape=(None,),
                dtype=self.dtype,
                compression=self.compression,
                compression_opts=self.compression_opts,
            )

    def has_blobs(self):
        with self.open() as f:
            return f[self.name].attrs["has_blobs"]

    def get_value(self, name, flat=False, thin=1, discard=0):
        if not self.initialized:
            raise AttributeError(
                "You must run the sampler with "
                "'store == True' before accessing the "
                "results"
            )
        with self.open() as f:
            g = f[self.name]
            iteration = g.attrs["iteration"]
            if iteration <= 0:
                raise AttributeError(
                    "You must run the sampler with "
                    "'store == True' before accessing the "
                    "results"
                )

            if name == "blobs" and not g.attrs["has_blobs"]:
                return None

            v = g[name][discard + thin - 1 : self.iteration : thin]
            if flat:
                s = list(v.shape[1:])
                s[0] = np.prod(v.shape[:2])
                return v.reshape(s)
            return v


    @property
    def shape(self):
        with self.open() as f:
            g = f[self.name]
            return g.attrs["nwalkers"], g.attrs["ndim"]

    @property
    def iteration(self):
        with self.open() as f:
            return f[self.name].attrs["iteration"]

    @property
    def accepted(self):
        with self.open() as f:
            return f[self.name]["accepted"][...]
        

    def grow(self, ngrow, blobs):
        """Expand the storage space by some number of samples

        Args:
            ngrow (int): The number of steps to grow the chain.
            blobs: The current array of blobs. This is used to compute the
                dtype for the blobs array.

        """
        self._check_blobs(blobs)

        with self.open("a") as f:
            g = f[self.name]
            ntot = g.attrs["iteration"] + ngrow
            g["chain"].resize(ntot, axis=0)
            g["log_like"].resize(ntot, axis=0)
            g["log_prob"].resize(ntot, axis=0)
            g["beta_history"].resize(ntot, axis=0)

            if blobs is not None:
                has_blobs = g.attrs["has_blobs"]
                if not has_blobs:
                    nwalkers = g.attrs["nwalkers"]
                    dt = np.dtype((blobs.dtype, blobs.shape[1:]))
                    g.create_dataset(
                        "blobs",
                        (ntot, nwalkers),
                        maxshape=(None, nwalkers),
                        dtype=dt,
                        compression=self.compression,
                        compression_opts=self.compression_opts,
                    )
                else:
                    g["blobs"].resize(ntot, axis=0)
                    if g["blobs"].dtype.shape != blobs.shape[1:]:
                        raise ValueError(
                            f'Existing blobs have shape {g["blobs"].dtype.shape} but new blobs requested with shape {blobs.shape[1:]}'
                        )
                g.attrs["has_blobs"] = True

    def save_step(self, state, accepted):
        """Save a step to the backend

        Args:
            state (State): The :class:`State` of the ensemble.
            accepted (ndarray): An array of boolean flags indicating whether
                or not the proposal for each walker was accepted.

        """
        self._check(state, accepted)

        with self.open("a") as f:
            g = f[self.name]
            iteration = g.attrs["iteration"]

            g["chain"][iteration, :, :] = state.coords
            g["log_like"][iteration, :] = state.log_like
            g["log_prob"][iteration, :] = state.log_prob
            g["beta_history"][iteration] = state.beta

            if state.blobs is not None:
                g["blobs"][iteration, :] = state.blobs
            g["accepted"][:] += accepted

            g.attrs["iteration"] = iteration + 1


    def swap_step(self, state):
         with self.open("a") as f:
            g = f[self.name]
            #iteration = g.attrs["iteration"]

            g["chain"][-1, :, :] = state.coords
            g["log_like"][-1, :] = state.log_like
            g["log_prob"][-1, :] = state.log_prob

            if state.blobs is not None:
                g["blobs"][-1, :] = state.blobs



class PTHDFBackend(object):
    def __init__(
        self,
        filename,
        name="mcmc",
        read_only=False,
        dtype=None,
        compression=None,
        compression_opts=None,
    ):
        if h5py is None:
            raise ImportError("you must install 'h5py' to use the HDFBackend")
        self.filename = filename
        self.filename_raw = filename.split('.')[0]
        self.name = name
        self.read_only = read_only
        self.compression = compression
        self.compression_opts = compression_opts
        self.backends = []
        if dtype is None:
            self.dtype_set = False
            self.dtype = np.float64
        else:
            self.dtype_set = True
            self.dtype = dtype

    
    @property
    def initialized(self):
        if not os.path.exists(self.filename):
            return False
        try:
            with self.open() as f:
                return self.name in f
        except (OSError, IOError):
            return False
        
    
    def open(self, mode="r"):
        if self.read_only and mode != "r":
            raise RuntimeError(
                "The backend has been loaded in read-only "
                "mode. Set `read_only = False` to make "
                "changes."
            )
        f = h5py.File(self.filename, mode)

        # TODO use another attr
        if not self.dtype_set and self.name in f:
            g = f[self.name]
            if "chain" in g:
                self.dtype = g["chain"].dtype
                self.dtype_set = True
        return f
    

    def reset(self, ntemps, nwalkers, ndim, tsw_hist=True, smd_hist=False):
        """Clear the state of the chain and empty the backend

        Args:
            ntemps (int): Number of temperatures.
            nwalkers (int): The size of the ensemble
            ndim (int): The number of dimensions

        """
        with self.open("a") as f:
            if self.name in f:
                del f[self.name]

            g = f.create_group(self.name)
            g.attrs["ntemps"] = ntemps
            g.attrs["nwalkers"] = nwalkers
            g.attrs["ndim"] = ndim
            g.attrs["has_blobs"] = False
            g.attrs["iteration"] = 0
            g.create_dataset(
                "ts_accepted",
                data=np.zeros(ntemps-1),
                compression=self.compression,
                compression_opts=self.compression_opts,
            )
            
            g.attrs['tsw_history_bool'] = tsw_hist
            g.attrs['smd_history_bool'] = smd_hist
            
            if tsw_hist:
                g.create_dataset(
                    "tsw_history",
                    (0, ntemps-1),
                    maxshape=(None, ntemps-1),
                    dtype=self.dtype,
                    compression=self.compression,
                    compression_opts=self.compression_opts,
                )

            if smd_hist:
                g.create_dataset(
                    "smd_history",
                    (0, ntemps-1),
                    maxshape=(None, ntemps-1),
                    dtype=self.dtype,
                    compression=self.compression,
                    compression_opts=self.compression_opts,
                )

            self.backends = [HDFBackend_plus(f'{self.filename_raw}_{t}.h5') for t in range(ntemps)]
            for backend in self.backends:
                backend.reset(nwalkers, ndim)


    def grow(self, ngrow):
        with self.open("a") as f:
            g = f[self.name]
            ntot = g.attrs["iteration"] + ngrow
            #tsw_history_bool = g["tsw_history_bool"]
            #smd_history_bool = g["smd_history_bool"]

            if self.tsw_history_bool:
                g["tsw_history"].resize(ntot, axis=0)

            if self.smd_history_bool:
                g["smd_history"].resize(ntot, axis=0)


    def save_step(self, states, ts_accepted, smd=None):
        """Save a step to the backend

        Args:
            state (State): The :class:`State` of the ensemble.
            accepted (ndarray): An array of boolean flags indicating whether
                or not the proposal for each walker was accepted.

        """
        for t, backend in enumerate(self.backends):
            backend.swap_step(states[t])


        with self.open("a") as f:
            g = f[self.name]
            iteration = g.attrs["iteration"]
            tsw_history_bool = g.attrs["tsw_history_bool"]
            smd_history_bool = g.attrs["smd_history_bool"]

            if tsw_history_bool:
                g["tsw_history"][iteration] = ts_accepted

            if smd_history_bool:
                g["smd_history"][iteration] = smd

            g["ts_accepted"][:] += ts_accepted
            g.attrs["iteration"] = iteration + 1
    


    def get_autocorr_time(self, discard=0, thin=1, **kwargs):
        return [b.get_autocorr_time(discard=discard, thin=thin, **kwargs) for b in self.backends]

    def get_log_prob(self, **kwargs):
        """Get log probabilities from all temperatures."""
        return self.get_value('log_prob', **kwargs)

    def get_log_like(self, **kwargs):
        """Get log likelihoods from all temperatures."""
        return self.get_value('log_like', **kwargs)

    def get_chain(self, **kwargs):
        """Get chains from all temperatures."""
        return self.get_value('chain', **kwargs)

    def get_betas(self, **kwargs):
        return self.get_value('beta_history', **kwargs)

    def get_value(self, name, **kwargs):
        return np.array([b.get_value(name, **kwargs) for b in self.backends])


    def get_tsw(self, **kwargs):
        return self.get_value_over('tsw_history', **kwargs)


    def get_smd(self, **kwargs):
        return self.get_value_over('smd_history', **kwargs)
        

    def get_value_over(self, name, flat=False, thin=1, discard=0):
        if not self.initialized:
            raise AttributeError(
                "You must run the sampler with "
                "'store == True' before accessing the "
                "results"
            )
        with self.open() as f:
            g = f[self.name]
            iteration = g.attrs["iteration"]
            if iteration <= 0:
                raise AttributeError(
                    "You must run the sampler with "
                    "'store == True' before accessing the "
                    "results"
                )

            if name == "blobs" and not g.attrs["has_blobs"]:
                return None

            v = g[name][discard + thin - 1 : self.iteration : thin]
            if flat:
                s = list(v.shape[1:])
                s[0] = np.prod(v.shape[:2])
                return v.reshape(s)
            return v


    @property
    def shape(self):
        """The dimensions of the ensemble `(ntemps, nwalkers, ndim)`."""
        with self.open() as f:
            g = f[self.name]
            return g.attrs["ntemps"], g.attrs["nwalkers"], g.attrs["ndim"]

    @property
    def accepted(self):
        """Acceptance arrays from all temperatures."""
        return np.array([b.accepted for b in self.backends])

    
    @property
    def random_state(self):
        with self.open() as f:
            elements = [
                v
                for k, v in sorted(f[self.name].attrs.items())
                if k.startswith("random_state_")
            ]
        return elements if len(elements) else None


    @property
    def tsw_history_bool(self):
        with self.open() as f:
            return f[self.name].attrs["tsw_history_bool"]


    @property
    def smd_history_bool(self):
        with self.open() as f:
            return f[self.name].attrs["smd_history_bool"]
        
    @property
    def iteration(self):
        with self.open() as f:
            return f[self.name].attrs["iteration"]
        

    def get_last_sample(self):
        """Access the most recent sample in the chain"""
        if (not self.initialized) or self.iteration <= 0:
            raise AttributeError(
                "you must run the sampler with "
                "'store == True' before accessing the "
                "results"
            )

        states = PTState([b.get_last_sample() for b in self.backends])
        states.random_state = self.random_state
        return states


    def __getitem__(self, n):
        return self.backends[n]


    def __iter__(self):
        return iter(self.backends)
    
    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        pass    


# @auto-fold regex /^\s*if/ /^\s*else/ /^\s*def/
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# my coding convention
# **EVAL : evaluate the performance of this method
# **RED  : redo this
# **DEB  : debugging needed in this part
# **DEL  : DELETE AT SOME POINT


class PTWrapper_1(object):
    """
    Make logl and logp pickable
    """

    def __init__(self, logl, logp, loglargs, loglkwargs, logpargs, logpkwargs):
        self.logl = logl
        self.logp = logp

        self.loglargs = loglargs or []
        self.loglkwargs = loglkwargs or {}
        self.logpargs = logpargs or []
        self.logpkwargs = logpkwargs or {}

    def __call__(self, x, beta):
        try:
            lp = self.logp(x, *self.logpargs, **self.logpkwargs)
            ll = self.logl(x, *self.loglargs, **self.loglkwargs)

            log_prob = lp + ll*beta

            return log_prob, ll
        
        except Exception as e:
            import traceback

            print("Exception while calling your likelihood function:")
            print("  params:", x)
            print("  exception:")
            traceback.print_exc()
            raise


class PTWrapper(object):
    """
    Make logl and logp pickable
    """

    def __init__(self, logl, logp, loglargs, loglkwargs, logpargs, logpkwargs):
        self.logl = logl
        self.logp = logp

        self.loglargs = loglargs or []
        self.loglkwargs = loglkwargs or {}
        self.logpargs = logpargs or []
        self.logpkwargs = logpkwargs or {}

    def __call__(self, x_beta):
        x, beta = x_beta
        
        lp = self.logp(x, *self.logpargs, **self.logpkwargs)
        if lp == float('-inf'):
            return lp, lp

        ll = self.logl(x, *self.loglargs, **self.loglkwargs)
        return lp + ll * beta, ll


# @auto-fold regex /^\s*if/ /^\s*else/ /^\s*def/
#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from .sampler import PTSampler
from .state import PTState
from .backend import PTBackend

__name__ = 'reddemcee'
__version__ = '0.9.10'
__url__ = "https://reddemcee.readthedocs.io"
__all__ = ['PTSampler', 'PTState', 'PTBackend']


# @auto-fold regex /^\s*if/ /^\s*else/ /^\s*def/
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# my coding convention
# **EVAL : evaluate the performance of this method
# **RED  : redo this
# **DEB  : debugging needed in this part
# **DEL  : DELETE AT SOME POINT


from itertools import product as iter_prod
from collections import namedtuple
from collections.abc import Iterable

from typing import Dict, List, Optional, Union

from .state import PTState
from .moves import StretchMove, PTMove
from .backend import PTBackend
from .wrapper import PTWrapper
from .utils import set_temp_ladder

from tqdm import tqdm
import numpy as np
from emcee import autocorr


class LadderAdaptation(object):
    def calc_decay(self):
        return self.config_adaptation_halflife / (self.backend[0].iteration + self.config_adaptation_halflife)

    def calc_dss(self):
        kappa = self.calc_decay() / self.config_adaptation_rate
        adjust = self.adjustment_fn()
        adjust[np.abs(adjust) < 1e-8] = 0   # Failsafe for very small values
        return kappa * adjust
    
    def select_adjustment(self, option=0):
        """Select an adjustment method.
        It tries to equalize the following:

        0: Temperature Swap Rate

        1: Swap Mean Distance

        2: Specific Heat

        3: dE/sig . Approximation propossed by...
        

        Args:
            option (int): Selects the method.
        """
        self.adjustment_fn = getattr(self, f'calc_adjust{option}')


    def calc_adjust00(self):
        # No adjustment
        return np.zeros(self.ntemps-2)

    def calc_adjust0(self):
        # Uniform Swap Acceptance Rate
        return -np.diff(self.ts_accepted)
    
    def calc_adjust1(self):
        # Swap Mean Distance
        return -np.diff(self.smd)
    
    def calc_adjust2(self):
        # SGG
        logls = self.get_log_like(flat=True)[:, -1000:]
        db = -np.diff(self.betas)
        varL = np.var(logls, axis=1)
        a = np.exp(-varL[:-1] * db**2)
        return -np.diff(a)

    def calc_adjust3(self):
        # dE_m/sig_m; rathore et al 2005
        # SAO
        logls = self.get_log_like(flat=True)[:, -1000:]
        Esig = np.std(logls, axis=1)
        Emean = np.mean(logls, axis=1)
        sig1 = Esig[:-1]
        sig2 = Esig[1:]
        sigmean = (sig1+sig2)/2
        diffE = np.diff(Emean)
        a = diffE/sigmean        
        return -np.diff(a)

    def calc_adjust4(self):
        '''THERMODYNAMIC LENGTH'''
        betas = np.asarray(self.betas)
        logls = self.get_log_like(flat=True)[:, -1000:]
        Esig = np.std(logls, axis=1)

        dL = -np.diff(betas) * (Esig[1:] + Esig[:-1]) / 2

        norm = -np.trapz(Esig, betas)/(len(betas)-1)
        dLn = dL/norm

        return np.diff(dLn)

    def calc_adjust5(self):
        E = np.mean(self.get_log_like(flat=True)[:, -1000:], axis=1)
        dE = np.diff(E)
        db = np.diff(self.betas)
        logA = -dE * db

        return -np.diff(logA)

    def calc_adjust6(self):
        # beta density
        betas = np.asarray(self.betas)[:-1]
        #t = 1/betas[:-1]
        #gamma = t[1:]/t[:-1]
        #nu = 1/np.log(gamma)
        #tmean = (t[:-1] - t[1:]) / (np.log(t[:-1]) - np.log(t[1:]))
        
        nu = -1/np.diff(np.log(betas))
        betamean = (betas[:-1] - betas[1:]) / (np.log(betas[:-1]) - np.log(betas[1:]))

        # CvL
        logls = self.get_log_like(flat=True)[:, -1000:]
        #Esig = np.std(logls, axis=1)[:-1]
        Esig = np.std(logls, axis=1)[:-1]

        Esigmean = (Esig[:-1] - Esig[1:]) / (np.log(Esig[:-1]) - np.log(Esig[1:]))
        #cvmean = Esigmean / tmean
        cvmean = Esigmean * betamean


        #nu_norm = -np.trapz(tmean, nu)
        #cv_norm = -np.trapz(tmean, sqrt_cvmean)
        nu_norm = np.sum(nu)
        cv_norm = np.sum(cvmean)
        #print(f'{betas=}')
        
        nu_normed = nu / nu_norm
        cv_normed = cvmean / cv_norm

        #print(f'{nu_normed=}')
        #print(f'{cv_normed=}')
        #print(f'coso = {(nu_normed-cv_normed)}')
        return (nu_normed-cv_normed)

    def calc_adjust7(self):
        # beta density
        betas = np.asarray(self.betas)
        betamean = (betas[:-1] + betas[1:]) / 2
        nub = -1/np.diff(np.log(betas))


        logls = self.get_log_like(flat=True)[:, -1000:]
        Esig = np.std(logls, axis=1)
        #Esigmean = (Esig[:-1] - Esig[1:]) / (np.log(Esig[:-1]) - np.log(Esig[1:]))
        Esigmean = (Esig[:-1] + Esig[1:]) / 2
        #cvmean = Esigmean / tmean
        cvmean = Esigmean * betamean


        nub_norm = np.sum(nub)
        cv_norm = np.sum(cvmean)

        a = (nub/nub_norm)-(cvmean/cv_norm)
        return -np.diff(a)

    def calc_adjust8(self):
        return -np.diff(np.log(self.ts_accepted))

    def calc_adjust9(self):
        # beta density
        betas = np.asarray(self.betas)
        logls = self.get_log_like(flat=True)[:, -1000:]
        Esig = np.std(logls, axis=1)

        dL = -np.diff(betas) * (Esig[1:] + Esig[:-1]) / 2

        norm = -np.trapz(Esig, betas)/(len(betas)-1)
        dLn = dL - norm
        print(f'{dL=}')
        print(f'{dLn=}')

        return np.diff(dLn)


class PTSampler(LadderAdaptation):
    def __init__(
        self,
        nwalkers,
        ndim,
        log_like,
        log_prior,
        betas=None,
        ntemps=None,
        pool=None,
        moves=None,
        loglargs=None,
        loglkwargs=None,
        logpargs=None,
        logpkwargs=None,
        backend=None,
        vectorize=False,
        blobs_dtype=None,
        smd_history=False,
        tsw_history=True,
        adapt_tau=1000,
        adapt_nu=1,
        adapt_mode=0,
        parameter_names: Optional[Union[Dict[str, int], List[str]]] = None,
    ):
        """Initialize the adaptive parallel tempering MCMC ensemble sampler.

        Args:
            nwalkers (int): The number of walkers in each ensemble.
            ndim (int): The number of dimensions in the parameter space.
            log_like (callable): The log-likelihood function.
            log_prior (callable): The log-prior function.
            betas (Optional[Iterable[float]]): The inverse temperatures of the parallel chains.
            ntemps (Optional[int]): The number of temperatures. If None, determined from betas.
            pool (Optional[Pool]): A pool object for parallel processing.
            moves (Optional[List[Move]]): A list of moves to use.
            loglargs (Optional[tuple]): Positional arguments for the log-likelihood function.
            loglkwargs (Optional[dict]): Keyword arguments for the log-likelihood function.
            logpargs (Optional[tuple]): Positional arguments for the log-prior function.
            logpkwargs (Optional[dict]): Keyword arguments for the log-prior function.
            backend (Optional[Backend]): A backend object for storing the chain.
            vectorize (Optional[bool]): Whether to vectorize the log-probability function.
            blobs_dtype (Optional[dtype]): The dtype of the blobs.
            smd_history (Optional[bool]): Whether to store swap mean distance history.
            tsw_history (Optional[bool]): Whether to store temperature swap history.
            adapt_tau (Optional[int]): Halflife of adaptation.
            adapt_nu (Optional[int]): Adaptation rate.
            adapt_mode (Optional[int]): Adaptation mode.
            parameter_names (Optional[Union[Dict[str, int], List[str]]]): Parameter names.

        """
        # Parse Move
        self._parse_moves(moves, tsw_history, smd_history)
        
        self.config_adaptation_halflife = adapt_tau
        self.config_adaptation_rate = adapt_nu
        self.config_adaptation_mode = adapt_mode
        self.select_adjustment(self.config_adaptation_mode)

        self.z_ngrid = 10001
        self.z_nsim = 1000 

        self.pool = pool
        self.vectorize = vectorize
        self.blobs_dtype = blobs_dtype

        self.ndim = ndim
        self.nwalkers = nwalkers
        self.ntemps = ntemps or len(betas)

        self.betas = betas or set_temp_ladder(ntemps, ndim)


        # Initialize random number generator
        self._random = np.random.RandomState()
        # Initialize the Backend
        self._init_backend(backend)

        # Wrap the log-probability functions
        self.log_prob_fn = PTWrapper(log_like, log_prior, loglargs, loglkwargs, logpargs, logpkwargs)

        # Save the parameter names
        self._parameter_names(parameter_names)


    def sample(
        self,
        initial_state, 
        nsteps=1,
        nsweeps=1,
        tune=False,
        thin_by=1,
        store=True,
        progress=False
    ):
        """Advance the chain as a generator

        Args:
            initial_state (State or ndarray[nwalkers, ndim]): The initial state of the walkers.
            nsteps (Optional[int]): The number of steps to generate.
            nsweeps (Optional[int]): The number of sweeps to generate.
            tune (Optional[bool]): If True, tune the parameters of some moves.
            thin_by (Optional[int]): Store every `thin_by` samples.
            store (Optional[bool]): If True, store the positions and log-probabilities.
            progress (Optional[bool or str]): Show a progress bar if True.

        Yields:
            State (State): The state of the ensemble at each step.

        """
        self._check_sample_init(nsteps, store, thin_by)
        # modify swap move in case of smd
        #self._swap_move.D_ = self.D_

        # Interpret the input as a walker state.
        state = PTState(initial_state, copy=True)

        # Check and Initialize states
        self._init_states(state)

        # Thin
        thin_by = int(thin_by)
        yield_step = checkpoint_step = thin_by

        iterations = int(nsteps * nsweeps)
        # Store
        if store:
            self.backend.grow(nsweeps)
            for t in range(self.ntemps):
                self.backend[t].grow(iterations, state[t].blobs)

        # Set up a wrapper around the relevant model functions
        map_fn = self.pool.map if self.pool is not None else map
        model = self._create_model(map_fn)

        # Inject the progress bar
        total = None if nsteps is None else iterations * yield_step * self.ntemps

        with tqdm(total=total, disable=not progress) as pbar:
            i = 0
            for _ in range(nsweeps):
                # SELECT SWAP FROM RANDOM CHOICE
                swap_move = self._swap_move
                state, self.ts_accepted, self.smd = swap_move.swap(state)

                for t in range(self.ntemps):
                    for _, _ in iter_prod(range(nsteps), range(yield_step)):
                        # Choose a random move
                        move = self._random.choice(self._moves, p=self._weights)

                        # Propose
                        state[t], accepted = move.propose(model, state[t])
                        state.random_state = self._random.get_state()

                        if tune:
                            move.tune(state[t], accepted)

                        # Save the new step
                        if store and (i + 1) % checkpoint_step == 0:
                            self.backend[t].save_step(state[t], accepted)

                        pbar.update(1)
                        i += 1

                # TEMPERATURE SWAP
                state, self.ts_accepted, self.smd = swap_move.swap(state)

                # TEMPERATURE LADDER ADAPTATION
                state = swap_move.adapt(state, self.calc_dss())

                # SAVE
                self.backend.save_step(state, self.ts_accepted, self.smd)

                # Overwrite sampler betas, this is for user
                self.betas = state.betas
                # Yield the result as an iterator
                yield state


    def run_mcmc(self, initial_state, nsteps, nsweeps, **kwargs):
        """
        Iterate :func:`sample` for ``nsweeps`` times ``nsteps`` iterations and return the result

        Args:
            initial_state (State or ndarray[nwalkers, ndim]): The initial state or position vector. Can also be
                ``None`` to resume from where :func:``run_mcmc`` left off the
                last time it executed.
            nsteps (int): The number of steps to run.
            nsweeps (int): The number of sweeps to run.


        Other parameters are directly passed to :func:`sample`.

        This method returns the most recent result from :func:`sample`.

        """
        if initial_state is None:
            if self._previous_state is None:
                raise ValueError(
                    "Cannot have `initial_state=None` if run_mcmc has never "
                    "been called."
                )
            initial_state = self._previous_state
        
        #self.ntemps__ = initial_state.shape[0]

        results = None
        for results in self.sample(initial_state,
                                   nsteps=nsteps,
                                   nsweeps=nsweeps,
                                   **kwargs):
            pass

        # Store so that the ``initial_state=None`` case will work
        self._previous_state = results

        return results


    def run_auto_mcmc(self, initial_state, maxiter,
                      init_steps=100, repeats=1, **kwargs):
        """
        Iterate :func:`sample` for ``nsweeps`` times ``nsteps`` iterations and return the result

        Args:
            initial_state (State or ndarray[nwalkers, ndim]): The initial state or position vector. Can also be
                ``None`` to resume from where :func:``run_mcmc`` left off the
                last time it executed.
            nsteps (int): The number of steps to run.
            nsweeps (int): The number of sweeps to run.


        Other parameters are directly passed to :func:`sample`.

        This method returns the most recent result from :func:`sample`.

        """
        if initial_state is None:
            if self._previous_state is None:
                raise ValueError(
                    "Cannot have `initial_state=None` if run_mcmc has never "
                    "been called."
                )
            initial_state = self._previous_state

        self.steps_per_sweep = []
        results = None
        ns_ = 0
        rp_ = 0
        act = init_steps
        while ns_ <= maxiter:
            for results in self.sample(initial_state,
                                    nsteps=act,
                                    nsweeps=repeats,
                                    **kwargs):
                pass
            ns_ += act
            self.steps_per_sweep.append(act)
            #if rp_ % repeats == 0:
            x = self.backend[0].get_log_like(discard=ns_//2)
            act = round(autocorr.integrated_time(x, tol=0)[0])
            #act = round(self.backend[0].get_autocorr_time(tol=0)))
            print(f'{ns_=} | {act=}')
            #if remaining == 1:
            #    self.select_adjustment(self.config_adaptation_mode)
            #    remaining = int(np.mean(self.backend[0].get_autocorr_time(tol=0)))
            #    print(f'{remaining=}')
            #else:
            #    self.select_adjustment('00')
            #    remaining -= 1
            # Store so that the ``initial_state=None`` case will work
            
            self._previous_state = results

        return results


    def compute_log_prob(self, coords, beta):
        """Calculate the vector of log-probability for the walkers.

        This method calculates the log-probability for each walker given their
        coordinates and the inverse temperature. It handles parameter naming,
        vectorization, and parallel processing.

        Args:
            coords (ndarray): The coordinates of the walkers.
            beta (float): The inverse temperature.

        Returns:
            Tuple[ndarray, ndarray, Optional[ndarray]]: A tuple containing the log
                probability, log likelihood, and blobs (if any).

        Raises:
            ValueError: If any parameter value is infinite or NaN, or if the
                probability function returns NaN.
        """
        p = coords

        # Check that the parameters are in physical ranges.
        if np.any(np.isinf(p)):
            raise ValueError("At least one parameter value was infinite")
        if np.any(np.isnan(p)):
            raise ValueError("At least one parameter value was NaN")

        if self.params_are_named:
            p = ndarray_to_list_of_dicts(p, self.parameter_names)

        beta_array = np.repeat(beta, len(p))

        if self.vectorize:
            results = self.log_prob_fn(zip(p, beta_array))
        else:
            map_func = self.pool.map if self.pool is not None else map
            results = list(map_func(self.log_prob_fn, zip(p, beta_array)))


        # Unpack results
        try:
            # Assume results are tuples: (log_prob, log_like, *blobs)
            log_prob = np.array([_scalar(l[0]) for l in results])
            log_like = np.array([_scalar(l[1]) for l in results])
            blobs = [l[2:] for l in results if len(l) > 2]
            blobs = self._process_blobs(blobs) if blobs else None
        except Exception as e:
            raise ValueError(f"Error in log_prob_fn: {e}") from e

        if np.any(np.isnan(log_prob)):
            raise ValueError("Probability function returned NaN")

        return log_prob, log_like, blobs


    def reset(self):
        """Reset the book-keeping parameters.

        This method resets the backend, clearing all stored data and resetting the
        iteration counter. It also re-initializes the backend with the current
        number of temperatures, walkers, dimensions, and swap history settings.
        """
        self.backend.reset(self.ntemps, self.nwalkers, self.ndim,
                           smd_hist=self._swap_move.smd_hist, tsw_hist=self._swap_move.tsw_hist)


    def get_evidence_ti(self, discard=0, thin=1, mode='obm',
                        ba=None, bb=None, pchip=True,
                        fixed_ladder=True, batch_size=None,
                        spe_kernel='bartlett'):

        L = self.get_log_like(discard=discard, thin=thin)  # shape=(ntemps, nsweeps, nwalkers)
        B = self.get_betas(discard=discard, thin=thin)  # shape=(ntemps, nsweeps)

        from .utils import get_ti
        return get_ti(B, L,
                      pchip=pchip,
                      ba=ba,
                      bb=bb,
                      fixed_ladder=fixed_ladder,
                      mode=mode,
                      batch_size=batch_size,
                      spe_kernel=spe_kernel)


    def get_evidence_ss(self, discard=0, thin=1, mode='obm',
                        ba=None, bb=None, bridge=True,
                        fixed_ladder=True, batch_size=None,
                        spe_kernel='bartlett'):
        
        L = self.get_log_like(discard=discard)  # shape=(ntemps, nsweeps, nwalkers)
        B = self.get_betas(discard=discard)  # shape=(ntemps, nsweeps)

        from .utils import get_ss
        return get_ss(B, L,
                      mode=mode,
                      ba=ba, bb=bb, bridge=bridge,
                      fixed_ladder=fixed_ladder, batch_size=batch_size,
                      spe_kernel=spe_kernel)


    def get_evidence_hybrid(self, discard=0, nbetacut=None,
                            nbetacut_crit='max',
                            ti_args=None, ss_args=None):
        if ti_args is None:
            ti_args = {'mode':'obm',
                       'ba':None, 'bb':None, 'pchip':True,
                       'fixed_ladder':True, 'batch_size':None,
                       'spe_kernel':'bartlett'}

        if ss_args is None:
            ss_args = {'mode':'obm',
                       'ba':None, 'bb':None, 'bridge':True,
                       'fixed_ladder':True, 'batch_size':None,
                       'spe_kernel':'bartlett'}


        L = self.get_log_like(discard=discard)  # shape=(ntemps, nsweeps, nwalkers)
        B = self.get_betas(discard=discard)  # shape=(ntemps, nsweeps)    

        

        if nbetacut is None:
            if nbetacut_crit == 'max':
                all_cuts = self._hybrid_max(B)
                        #print(f'{max_index=}')
            elif nbetacut_crit=='tri':
                all_cuts = self._hybrid_tri(B, L)
                        #print(max_index)
            else:
                print('MODE NOT DETECTED')
        else:
            all_cuts = self._hybrid_def(nbetacut)
        
        ti_args['ba'] = all_cuts[0]
        ti_args['bb'] = all_cuts[1]

        ss_args['ba'] = all_cuts[2]
        ss_args['bb'] = all_cuts[3]

        from .utils import get_ss, get_ti

        zti, ztierr = get_ti(B, L, **ti_args)

        zss, zsserr = get_ss(B, L, **ss_args)

        z_hat = zti + zss
        err_tot = np.sqrt(zsserr**2 + ztierr**2)

        return z_hat, err_tot




    def _hybrid_max(self, B):
        betas = B[::-1, -1]
        if betas[0] == 0:
            betas[0] = np.finfo(np.float32).eps
        y = 1/np.diff(np.log(betas))
        max_index = np.argmax(y)
        return self._hybrid_def(max_index)

    def _hybrid_tri(self, B, L):
        x = B[::-1, -1]
        y = np.mean(np.mean(L, axis=2), axis=1)[::-1]

        cuad = np.diff(x) * y[:-1]
        triang = np.diff(x) * np.diff(y)/2
        mask = triang<=cuad

        max_index = np.sum(mask)
        return self._hybrid_def(max_index)

    def _hybrid_def(self, nbetacut):
        ba_ti = 0
        bb_ti = nbetacut
        ba_ss = nbetacut
        bb_ss = self.ntemps - 1

        return ba_ti, bb_ti, ba_ss, bb_ss




    def get_autocorr_time(self, **kwargs):
        """Get the estimated autocorrelation time.

        This method returns the estimated autocorrelation time from the backend. It is useful
        for assessing the convergence of the chains.

        Args:
            **kwargs (Optional[dict]): Additional keyword arguments passed to the backend.

        Returns:
            (ndarray): The estimated autocorrelation time.
        """
        return self.backend.get_autocorr_time(**kwargs)

    def get_betas(self, **kwargs):
        """Get the betas of the chains.

        This method returns the betas history of the chains from the backend. It is useful
        for checking the temperature ladder adaptation.

        Args:
            **kwargs (Optional[dict]): Additional keyword arguments passed to the backend.

        Returns:
            (ndarray): The betas of the chains.
        """
        return self.get_value("beta_history", **kwargs)

    def get_chain(self, **kwargs):
        """Get the chain of samples.

        This method returns the chain of samples from the backend. It is useful
        for analyzing the posterior distribution.

        Args:
            **kwargs (Optional[dict]): Additional keyword arguments passed to the backend.

        Returns:
            (ndarray): The chain of samples.
        """
        return self.get_value("chain", **kwargs)

    def get_blobs(self, **kwargs):
        """Get the blobs.

        This method returns the blobs from the backend. Blobs are extra information
        returned by the log-probability function.
        Args:
            **kwargs (Optional[dict]): Additional keyword arguments passed to the backend.

        Returns:
            (ndarray): The chain of samples.
        """
        return self.get_value("blobs", **kwargs)

    def get_log_prob(self, **kwargs):
        """Get the log probability.

        This method returns the log probability from the backend. It is useful for
        analyzing the convergence of the chains.

        Args:
            **kwargs (Optional[dict]): Additional keyword arguments passed to the backend.

        Returns:
            (ndarray): The log probability.
        """
        return self.get_value("log_prob", **kwargs)

    def get_log_like(self, **kwargs):
        """Get the log likelihood.

        This method returns the log likelihood from the backend. It is useful for
        analyzing the posterior distribution.

        Args:
            **kwargs (Optional[dict]): Additional keyword arguments passed to the backend.

        Returns:
            (ndarray): The log likelihood.
        """
        return self.get_value("log_like", **kwargs)

    def get_last_sample(self, **kwargs):
        """Get the last sample.

        This method returns the last sample from the backend. It is useful for
        checking the current state of the chains.

        Args:
            **kwargs (Optional[dict]): Additional keyword arguments passed to the backend.

        Returns:
            (State): The last sample.
        """
        return self.backend.get_last_sample()

    def get_value(self, name, **kwargs):
        """Get a value from the backend.

        This method returns a value from the backend by its name. It is useful for
        accessing specific data stored during sampling.

        Args:
            name (str): The name of the value to retrieve.
            **kwargs (Optional[dict]): Additional keyword arguments passed to the backend.

        Returns:
            The value from the backend.
        """
        return self.backend.get_value(name, **kwargs)


    def _parse_moves(self, moves, tsw_history, smd_history):
        """Parse and initialize the moves"""
        if moves is None:
            self._moves = [StretchMove()]
            self._weights = [1.0]
        elif isinstance(moves, Iterable):
            # Check if moves is a sequence of (move, weight) tuples
            first_elem = next(iter(moves))
            if isinstance(first_elem, tuple) and len(first_elem) == 2:
                self._moves, self._weights = zip(*moves)
            else:
                self._moves = list(moves)
                self._weights = np.ones(len(self._moves))
        else:
            self._moves = [moves]
            self._weights = [1.0]

        self._weights = np.array(self._weights, dtype=float)
        self._weights /= np.sum(self._weights)

        # Parse the APT move schedule
        self._swap_move = PTMove()
        if tsw_history:
            self._swap_move.tsw_hist = True

        if smd_history:
            self._swap_move.smd_hist = True

    def _process_blobs(self, blobs):
        """Process blobs to ensure consistent dtype and shape"""
        if self.blobs_dtype is not None:
            dt = self.blobs_dtype
        else:
            try:
                dt = np.array(blobs[0]).dtype
            except Exception:
                dt = np.dtype('object')
        return np.array(blobs, dtype=dt)

    def _check_sample_init(self, nsteps, store, thin_by):
        if nsteps is None and store:
            raise ValueError("'store' must be False when 'nsteps' is None")
        if thin_by <= 0:
            raise ValueError("Invalid thinning argument")

    def _check_states(self, states):
        states_shape = np.shape(states)
        if states_shape != (self.ntemps, self.nwalkers, self.ndim):
            raise ValueError(f"incompatible input dimensions {states_shape}")
        
        for st in states:
            if not walkers_independent(st.coords):
                raise ValueError(
                    "Initial state has a large condition number. "
                    "Make sure that your walkers are linearly independent for the "
                    "best performance"
                )

    def _init_states(self, states):
        # Check the dimensions and walker independence
        self._check_states(states)

        for t in range(self.ntemps):
            state = states[t]
            beta = self.betas[t]

            if state.log_prob is None:
                state.beta = beta
                state.log_prob, state.log_like, state.blobs = self.compute_log_prob(state.coords, beta)

            if np.shape(state.log_prob) != (self.nwalkers,):
                raise ValueError("incompatible input dimensions")
            if np.any(np.isnan(state.log_prob)):
                raise ValueError("The initial log_prob was NaN")        

    def _create_model(self, map_fn):
        """Create a model for log probability calculations."""
        return namedtuple("Model", ("log_prob_fn", "compute_log_prob_fn", "map_fn", "random"))(
            self.log_prob_fn, self.compute_log_prob, map_fn, self._random
        )

    def _init_backend(self, backend):

        self.backend = PTBackend() if backend is None else backend
        if not self.backend.initialized:
            self._previous_state = None
            self.backend.reset(self.ntemps, self.nwalkers, self.ndim,
                               smd_hist=self._swap_move.smd_hist, tsw_hist=self._swap_move.tsw_hist)

            rstate = np.random.get_state()
            self.backend.random_state = rstate  # MARK1
        else:
            # TODO check previous backend shape?

            rstate = self.backend.random_state
            if rstate is None:
                rstate = np.random.get_state()

            # Grab the last step so that we can restart
            it = self.backend.iteration
            if it > 0:
                self._previous_state = self.get_last_sample()

        self._random.set_state(rstate)  # MARK1

    def _parameter_names(self, parameter_names):
        self.params_are_named: bool = parameter_names is not None
        if self.params_are_named:
            assert isinstance(parameter_names, (list, dict))
            assert not self.vectorize, "Named parameters with vectorization unsupported for now"

            if isinstance(parameter_names, list):
                assert len(parameter_names) == self.ndim, "Name all parameters or set `parameter_names` to `None`"
                parameter_names = {name: i for i, name in enumerate(parameter_names)}

            values = [
                v if isinstance(v, list) else [v]
                for v in parameter_names.values()
            ]
            values = {item for sublist in values for item in sublist}
            assert values == set(range(self.ndim)), f"Not all values appear -- set should be 0 to {self.ndim-1}"
            self.parameter_names = parameter_names

    @property
    def iteration(self):
        return self.backend.iteration


    @property
    def acceptance_fraction(self):
        """The fraction of proposed steps that were accepted."""
        return self.backend.accepted / float(self.backend[0].iteration)


    #@property
    def get_tsw(self, **kwargs):
        #temperature_swap_fraction
        """The fraction of proposed swaps that were accepted."""
        return self.backend.get_tsw(**kwargs)


    #@property
    def get_smd(self, **kwargs):
        # swap_mean_distance
        """The swap mean distance, normalised."""
        return self.backend.get_smd(**kwargs)


    def __getstate__(self):
        # In order to be generally picklable, we need to discard the pool
        # object before trying.
        d = self.__dict__.copy()
        d["pool"] = None
        return d



def walkers_independent(coords):
    if not np.all(np.isfinite(coords)):
        return False
    C = coords - np.mean(coords, axis=0)[None, :]
    C_colmax = np.amax(np.abs(C), axis=0)
    if np.any(C_colmax == 0):
        return False
    C /= C_colmax
    C_colsum = np.sqrt(np.sum(C**2, axis=0))
    C /= C_colsum
    return np.linalg.cond(C.astype(float)) <= 1e8


def ndarray_to_list_of_dicts(
    x: np.ndarray, key_map: Dict[str, Union[int, List[int]]]
) -> List[Dict[str, Union[np.number, np.ndarray]]]:
    """
    A helper function to convert a ``np.ndarray`` into a list
    of dictionaries of parameters. Used when parameters are named.

    Args:
      x (np.ndarray): parameter array of shape ``(N, n_dim)``, where
        ``N`` is an integer
      key_map (Dict[str, Union[int, List[int]]):

    Returns:
      list of dictionaries of parameters
    """
    return [{key: xi[val] for key, val in key_map.items()} for xi in x]


def _scalar(fx):
    # Make sure a value is a true scalar
    # 1.0, np.float64(1.0), np.array([1.0]), np.array(1.0)
    if not np.isscalar(fx):
        try:
            fx = np.asarray(fx).item()
        except (TypeError, ValueError) as e:
            raise ValueError("log_prob_fn should return scalar") from e
    return float(fx)


# @auto-fold regex /^\s*if/ /^\s*else/ /^\s*def/
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# my coding convention
# **EVAL : evaluate the performance of this method
# **RED  : redo this
# **DEB  : debugging needed in this part
# **DEL  : DELETE AT SOME POINT

import numpy as np

def thermodynamic_integration_classic(betas, logls):
    """Calculate thermodynamical integration according to Vousden et al """
    if betas[-1] != 0:
        betas1 = np.concatenate((betas, [0]))
        betas2 = np.concatenate((betas[::2], [0]))
        logls1 = np.concatenate((logls, [logls[-1]]))
        logls2 = np.concatenate((logls[::2], [logls[-1]]))
    else:
        betas1 = betas
        betas2 = np.concatenate((betas1[:-1:2], [0]))
        logls1 = logls
        logls2 = np.concatenate((logls1[:-1:2], [logls1[-1]]))

    logZ1 = -np.trapz(logls1, betas1)
    logZ2 = -np.trapz(logls2, betas2)

    return logZ1, np.abs(logZ1 - logZ2)


def thermodynamic_integration_upd(betas, logls1, ngrid=10001, nsim=1001):
    """Calculates the evidence (Z) and its associated error using interpolation techniques. This method employs the PCHIP (Piecewise Cubic Hermite Interpolating Polynomial) method to estimate log-likelihood values across a specified range of beta values, incorporating random noise to simulate variability.

    Args:
        **kwargs: Additional keyword arguments passed to the log-likelihood function.

    Returns:
        tuple: A tuple containing:
            - integral_trapz (float): The calculated evidence (Z).
            - err (float): The total error associated with the evidence.
            - err_disc (float): The discretization error.
            - err_samp (float): The sampling error.
    """
    from emcee.autocorr import integrated_time
    # Constants defined for this evaluation
    num_grid = ngrid
    num_simulations = nsim

    x = betas
    # Get LogL Means
    y = np.mean(logls1, axis=1)

    # Support Variables
    # TODO extract this method? autocorr outside
    n = np.array([len(l) for l in logls1])
    tau = np.array([integrated_time(logls1[i], c=5, tol=5, quiet=False)
                    for i in range(len(y))]).flatten()

    neff = n/tau
    stdyn = np.std(logls1, axis=1, ddof=1) / np.sqrt(neff)

    # Get Interpolated Values
    xnew, mean_interpolated, std_interpolated = interpolate(x, y, stdyn, num_grid, num_simulations)
    # Calculate Z
    integral_trapz = np.trapz(mean_interpolated, xnew)

    # Calculate Errors
    z_classic, zerr_classic = thermodynamic_integration_classic(x[::-1], y[::-1])
    err_disc = abs(z_classic - integral_trapz)

    h = np.diff(xnew)
    err_samp = np.sum((h/2)**2 * (std_interpolated[1:]**2 +
                                  std_interpolated[:-1]**2))
    #err_samp = 1/num_grid**2 * (1/4*std_interpolated[0]**2 +
    #                            np.sum(std_interpolated[1:-1]**2) +
    #                            1/4*std_interpolated[-1]**2)
    err_samp = np.sqrt(err_samp)
    err = np.sqrt(err_disc**2+err_samp**2)

    return integral_trapz, err, err_disc, err_samp


def interpolate(x, y, stdyn, num_grid, num_simulations):
    from scipy.interpolate import PchipInterpolator
    xnew = np.linspace(min(x), max(x), num_grid)
    
    # Interpolated values storage
    interpolated_values = np.zeros((num_simulations, num_grid))

    for i in range(num_simulations):
        # Perturb data with random noise based on errors
        y_noisy = y + np.random.normal(0, stdyn)

        # Create PchipInterpolator object with noisy data
        pchip = PchipInterpolator(x, y_noisy)

        # Interpolate over fine grid and store results
        interpolated_values[i, :] = pchip(xnew)
    
    mean_interpolated = np.mean(interpolated_values, axis=0)
    std_interpolated = np.std(interpolated_values, axis=0)
    return xnew, mean_interpolated, std_interpolated



def stepping_stones_classic(x1, y1):
    pass






def interpolate_future(x, y, stdyn, num_grid, num_simulations):
    from scipy.interpolate import PchipInterpolator
    minx = min(x) or np.finfo(np.float32).eps
    #xnew = np.linspace(minx, max(x), num_grid)
    xnew = np.geomspace(minx, max(x), num_grid)
    h = np.diff(xnew)
    # Interpolated values storage
    interpolated_values = np.zeros((num_simulations, num_grid))
    interpolated_error = np.zeros(num_simulations)


    for i in range(num_simulations):
        # Perturb data with random noise based on errors
        y_noisy = np.random.normal(y, stdyn)

        # Create PchipInterpolator object with noisy data
        pchip = PchipInterpolator(x, y_noisy)
        pchip_deriv = pchip.derivative()

        # Interpolate over fine grid and store results
        interpolated_values[i, :] = pchip(xnew)
        fprime_a = pchip_deriv(xnew)
        # Vectorised finite differences to approximate the second derivative over each interval
        diff_fprime = np.diff(fprime_a)      # f'(x[j+1]) - f'(x[j]) for each interval
        fsecond = diff_fprime / h          # Approximate second derivative on each interval
        
        err_intervals = - (h**3 / 12) * fsecond
        interpolated_error[i] = np.sum(np.abs(err_intervals))
    
    
    mean_interpolated = np.mean(interpolated_values, axis=0)
    std_interpolated = np.std(interpolated_values, axis=0)
    return xnew, mean_interpolated, std_interpolated, interpolated_error

def thermodynamic_integration(betas, logls, ngrid=101, nsim=1001):
    from emcee.autocorr import integrated_time
    # Constants defined for this evaluation
    num_grid = ngrid
    num_simulations = nsim

    x = betas
    yy = logls
    # Get LogL Means
    y = np.mean(yy, axis=1)

    n = np.array([len(l) for l in yy])
    tau = np.array([integrated_time(yy[i],
                                    c=5,
                                    tol=5,
                                    quiet=False)
                    for i in range(len(y))]).flatten()
    neff = n/tau
    stdyn = np.std(yy, axis=1, ddof=1) / np.sqrt(neff)
    xnew, mean_interpolated, std_interpolated, error_interpolated = interpolate_future(x, y, stdyn, num_grid, num_simulations)

    # Calculate Z
    integral_trapz = np.trapz(mean_interpolated, xnew)

    # Calculate Errors
    z_classic, zerr_classic = thermodynamic_integration_classic(x[::-1], y[::-1])
    err_disc = abs(z_classic - integral_trapz)

    err_disc = np.mean(error_interpolated)

    err_samp = 1/num_grid**2 * (1/4*std_interpolated[0]**2 +
                                np.sum(std_interpolated[1:-1]**2) +
                                1/4*std_interpolated[-1]**2)
    
    h = np.diff(xnew)
    err_samp = np.sum((h/2)**2 * (std_interpolated[1:]**2+std_interpolated[:-1]**2))


    err_samp = np.sqrt(err_samp)
    err = np.sqrt(err_disc**2+err_samp**2)

    return integral_trapz, err, err_disc, err_samp


def stepping_stones(x1, y1,
              nb_blocks=10):

    unique_betas = np.unique(x1)
    logZ = 0.0
    variance_components = []
    # We form stepping stones from temperature index 1 to ntemps-1.
    for i in range(1, len(unique_betas)):
        # Stepping stone: from beta_{i-1} to beta_i.
        delta_beta = unique_betas[i] - unique_betas[i-1]
        # Select all samples corresponding to the current beta level.
        mask = (unique_betas[i-1] <= x1) & (x1 < unique_betas[i])
        
        # Get likelihoods for temperature i; shape: (nsweeps, nwalkers)
        # For each sweep, average over walkers to get a weight
        # (this produces a time series of length nsweeps).
        weights = np.exp(delta_beta * y1[mask])
        term = np.mean(weights)
        # Full estimate: average over all sweeps.
        logZ += np.log(term)

        # Divide the sweeps into nb_blocks (using nearly equal sizes)
        blocks = np.array_split(weights.flatten(), nb_blocks)
        jackknife_estimates = []

        for b in range(nb_blocks):
            # Combine all blocks except block b.
            remaining = np.concatenate([blocks[j] for j in range(nb_blocks) if j != b])
            if remaining.size == 0:
                raise ValueError("Block size or number of blocks is inappropriate.")
            jk_mean = np.mean(remaining)
            jackknife_estimates.append(np.log(jk_mean))
        
        jackknife_estimates = np.array(jackknife_estimates)
        # Standard jackknife variance estimate.
        jk_mean = np.mean(jackknife_estimates)
        # Variance for this stepping stone.
        var_i = (nb_blocks - 1) / nb_blocks * np.sum((jackknife_estimates - jk_mean)**2)
        variance_components.append(var_i)

    # Assuming independence of contributions from different stepping stones,
    # the total variance is the sum of the individual variances.
    total_variance = np.sum(variance_components)
    err_logZ = np.sqrt(total_variance)
    
    return logZ, err_logZ


temp_table = np.array([25.2741, 7., 4.47502, 3.5236, 3.0232,
                      2.71225, 2.49879, 2.34226, 2.22198, 2.12628,
                      2.04807, 1.98276, 1.92728, 1.87946, 1.83774,
                      1.80096, 1.76826, 1.73895, 1.7125, 1.68849,
                      1.66657, 1.64647, 1.62795, 1.61083, 1.59494,
                      1.58014, 1.56632, 1.55338, 1.54123, 1.5298,
                      1.51901, 1.50881, 1.49916, 1.49, 1.4813,
                      1.47302, 1.46512, 1.45759, 1.45039, 1.4435,
                      1.4369, 1.43056, 1.42448, 1.41864, 1.41302,
                      1.40761, 1.40239, 1.39736, 1.3925, 1.38781,
                      1.38327, 1.37888, 1.37463, 1.37051, 1.36652,
                      1.36265, 1.35889, 1.35524, 1.3517, 1.34825,
                      1.3449, 1.34164, 1.33847, 1.33538, 1.33236,
                      1.32943, 1.32656, 1.32377, 1.32104, 1.31838,
                      1.31578, 1.31325, 1.31076, 1.30834, 1.30596,
                      1.30364, 1.30137, 1.29915, 1.29697, 1.29484,
                      1.29275, 1.29071, 1.2887, 1.28673, 1.2848,
                      1.28291, 1.28106, 1.27923, 1.27745, 1.27569,
                      1.27397, 1.27227, 1.27061, 1.26898, 1.26737,
                      1.26579, 1.26424, 1.26271, 1.26121,
                      1.25973])


def set_temp_ladder(ntemps, ndims, temp_table=temp_table):
    tstep = temp_table[ndims-1]
    if ndims > len(temp_table):
        tstep = 1.0 + 2.0*np.sqrt(np.log(4.0))/np.sqrt(ndims)
        
    return np.geomspace(1, tstep**(1-ntemps), ntemps)


# @auto-fold regex /^\s*if/ /^\s*else/ /^\s*def/
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# my coding convention
# **EVAL : evaluate the performance of this method
# **RED  : redo this
# **DEB  : debugging needed in this part
# **DEL  : DELETE AT SOME POINT

import numpy as np

# OBM helpers

def _obm_univariate(y, batch_size=None):
    """
    Univariate OBM estimator of the long-run variance Sig for
    a stationary scalar time series {y_t}.
    Returns Sig so that Var(y) ~ Sig / nsweeps
    """
    y = np.asarray(y, dtype=float)
    T = y.size
    if T < 3:
        raise ValueError('Need at least 3 sweeps for OBM')
    b = int(np.floor(T**0.5)) if batch_size is None else int(batch_size)
    b = max(2, min(b, T - 1))
    K = T - b + 1

    mu = y.mean()
    csum = np.concatenate(([0.0], np.cumsum(y)))
    bm = (csum[b:] - csum[:-b]) / b
    diff = bm - mu
    # long-run variance estimator
    Sigma_hat = (b / K) * np.dot(diff, diff)
    return Sigma_hat


def _obm_multivariate(Y, batch_size=None):
    """
    Multivariate OBM estimator for the long-run covariance Sig
    of a vector time series Y (shape: nsweeps, ntemps).
    Returns Sig so Var(Y) ≈ Sig / nsweeps
    """
    Y = np.asarray(Y, dtype=float)
    T, p = Y.shape
    if T < 3:
        raise ValueError('Need at least 3 sweeps for OBM')
    b = int(np.floor(T**0.5)) if batch_size is None else int(batch_size)
    b = max(2, min(b, T - 1))
    K = T - b + 1

    mu_hat = Y.mean(axis=0)
    csum = np.vstack([np.zeros((1, p)), np.cumsum(Y, axis=0)])
    bm = (csum[b:] - csum[:-b]) / b
    diff = bm - mu_hat
    Sigma_hat = (b / K) * diff.T @ diff
    return Sigma_hat


# Spectral helpers
def _autocovariances(y, maxlag):
    y = np.asarray(y, dtype=float)
    T = y.size
    mu = y.mean()
    z = y - mu
    gammas = np.empty(maxlag + 1, dtype=float)
    gammas[0] = np.dot(z, z) / T
    for k in range(1, maxlag + 1):
        gammas[k] = np.dot(z[k:], z[:-k]) / T
    return gammas

def _kernel_weights(kernel, b):
    k = np.arange(1, b + 1, dtype=float)
    if kernel == "bartlett":  # Newey–West
        w = 1.0 - k / (b + 1.0)
    elif kernel == "parzen":
        x = k / (b + 0.0)
        w = np.where(
            x <= 0.5,
            1 - 6*x**2 + 6*x**3,
            2*(1 - x)**3
        )
        w[x > 1] = 0.0
    elif kernel in {"tukey", "hann", "tukey-hanning", "tukeyhanning"}:
        # Cosine (Hann) lag window: w_b = 0, w_0 = 1
        w = 0.5 * (1 + np.cos(np.pi * k / (b + 0.0)))
    else:
        raise ValueError(f"Unknown kernel '{kernel}'")
    w[w < 0] = 0.0
    return w

def _spe_univariate(y, bandwidth=None, kernel="bartlett", prewhiten=False):
    """
    Estimate long-run variance Sig = f(0) via HAC (spectral density at zero).
    Var(y) ~ Sig / T.
    - kernel: 'bartlett' (Newey–West), 'parzen', or 'tukey' (Hann).
    - bandwidth: max lag b; if None, auto b = max(2, floor(2*T**(1/3))).
    - prewhiten: simple AR(1) prewhitening
    """
    y = np.asarray(y, dtype=float)
    T = y.size
    if T < 3:
        raise ValueError("Need at least 3 observations for spectral HAC.")

    if prewhiten:
        y0 = y - y.mean()
        num = np.dot(y0[1:], y0[:-1])
        den = np.dot(y0, y0)
        rho = 0.0 if den == 0 else np.clip(num / den, -0.97, 0.97)
        v = y[1:] - rho * y[:-1]
        Sigma_v = _spe_univariate(v, bandwidth=bandwidth, kernel=kernel, prewhiten=False)
        return Sigma_v / (1.0 - rho)**2

    b = int(np.floor(2 * T**(1/3))) if bandwidth is None else int(bandwidth)
    b = max(2, min(b, T - 1))

    gam = _autocovariances(y, b)
    w = _kernel_weights(kernel, b)
    Sigma = gam[0] + 2.0 * np.dot(w, gam[1:])

    return float(max(Sigma, 0.0))

def _spe_multivariate(Y, bandwidth=None, kernel="bartlett"):
    """
    Multivariate HAC estimator of the long-run covariance
    matrix Sig for a vector time series Y (nsweeps, ntemps).

    Returns Sig so Var(Y) ~ Sig/ nsweeps
    """
    Y = np.asarray(Y, dtype=float)
    if Y.ndim != 2:
        raise ValueError("Y must be 2D: (nsweeps, nsteps).")
    T, p = Y.shape
    if T < 3:
        raise ValueError("Need at least 3 sweeps for spectral HAC.")

    b = int(np.floor(2 * T**(1/3))) if bandwidth is None else int(bandwidth)
    b = max(2, min(b, T - 1))

    Z = Y - Y.mean(axis=0, keepdims=True)
    Gam0 = (Z.T @ Z) / T

    
    weights = _kernel_weights(kernel, b)
    Sigma = Gam0.copy()
    for k, wk in enumerate(weights, start=1):
        Zk = Z[k: , :]
        Z0 = Z[:-k, :]
        Gamk = (Zk.T @ Z0) / T
        Sigma += wk * (Gamk + Gamk.T)

    # numerical symmetrization, better stability
    Sigma = 0.5 * (Sigma + Sigma.T)
    return Sigma


# general helpers

def order_by_betas(B, L):
    B_sorted = B[::-1]
    L_sorted = L[::-1, :, :]
    return B_sorted, L_sorted

def _coarse_rows(a):
    if a.size == 0:
        return a
    base = a[::2, :]
    return base if len(a) % 2 == 1 else np.vstack([base, a[-1, :]])

def _coarse_rows_drop_hot(a):
    if a.size == 0:
        return a
    base = a[::2, :]
    return base if len(a) % 2 == 1 else a[1::2, :]

def _logmeanexp(x, axis=None):
    # numerical stability
    a = np.max(x, axis=axis, keepdims=True)
    return (a + np.log(np.mean(np.exp(x - a), axis=axis, keepdims=True))).squeeze()


# actual utility
def get_act_from_obm(y, batch_size=None, ddof=1):
    Sigma = _obm_univariate(y, batch_size)
    s2 = np.var(y, ddof=ddof)
    tau = Sigma / s2
    ESS = y.size / tau
    return tau, ESS


# pchip helper
def _pchip_integral_series_cut(B, L, cut_low, cut_high):
    """
    Per-sweep PCHIP integrals using the *full ladder* but
    integrating from cut_low to cut_high.

    Returns the evidence integral.
    """
    from scipy.interpolate import PchipInterpolator
    ntemps, nsweeps = B.shape
    z = np.zeros(nsweeps, dtype=float)

    for t in range(nsweeps):
        bt = B[:, t]
        ut = L[:, t]
        # collapse duplicates
        xu, inv = np.unique(bt, return_inverse=True)
        if xu.size != bt.size:
            uu = np.zeros_like(xu)
            for k in range(xu.size):
                uu[k] = ut[inv == k].mean()
            bt, ut = xu, uu

        if bt.size < 2:
            z[t] = 0.0
            continue

        # clamp cut index to valid segment range
        #ci = min(max(cut_index, 0), bt.size - 1)
        a = max(cut_low, bt[0])
        b = min(cut_high, bt[-1])
        if a >= b:
            z[t] = 0.0
            continue

        f = PchipInterpolator(bt, ut, extrapolate=False)
        z[t] = float(f.integrate(a, b))  # positive integral
    #print(a, b)
    return z



# SS helpers
def _ss_classic(B, L, fixed_ladder):
    ntemps, nsweeps = B.shape

    # Set up for SS
    m = ntemps-1
    W_series = np.empty((nsweeps, m), dtype=float)
    # Sampling Error
    if fixed_ladder:
        delta = (B[1:, :].mean(axis=1) - B[:-1, :].mean(axis=1))  # (m,)
        for i in range(m):  # ratio i uses base temperature index i
            s = delta[i] * L[i, :, :]  # (nsweeps, nwalkers)
            a = np.max(s, axis=1, keepdims=True)  # log-sum-exp stabilizer
            W_series[:, i] = np.exp(a).ravel() * np.mean(np.exp(s - a), axis=1)
    else:
        # Use per-sweep delta beta
        for i in range(m):
            dbeta_t = B[i+1, :] - B[i, :]  # (nsweeps,)
            s = dbeta_t[:, None] * L[i, :, :]  # (nsweeps, nwalkers)
            a = np.max(s, axis=1, keepdims=True)
            W_series[:, i] = np.exp(a).ravel() * np.mean(np.exp(s - a), axis=1)

    mu_hat = W_series.mean(axis=0)
    z_hat = np.sum(np.log(mu_hat))

    # grad
    grad = 1.0 / mu_hat  # for var

    return z_hat, W_series, grad


def _ss_bridge(B, L, fixed_ladder):
    ntemps, nsweeps = B.shape

    # Set up for SS
    m = ntemps-1

    logA_series = np.empty((nsweeps, m), dtype=float)
    logC_series = np.empty((nsweeps, m), dtype=float)

    # Sampling Error
    if fixed_ladder:
        # Robust constant delta-beta (average in case of tiny jitter)
        delta = (B[1:, :].mean(axis=1) - B[:-1, :].mean(axis=1))  # (m,)
        half = 0.5 * delta
        for i in range(m):
            # per-sweep log-mean-exp across walkers delta-beta i, i+1
            sA = half[i] * L[i, :, :]  # (nsweeps, nwalkers)
            sC = -half[i] * L[i+1, :, :]
            logA_series[:, i] = _logmeanexp(sA, axis=1)
            logC_series[:, i] = _logmeanexp(sC, axis=1)
    else:
        for i in range(m):
            dbeta_t = B[i+1, :] - B[i, :]  # (nsweeps,)
            sA = (0.5 * dbeta_t)[:, None] * L[i, :, :]
            sC = (-0.5 * dbeta_t)[:, None] * L[i+1, :, :]
            logA_series[:, i] = _logmeanexp(sA, axis=1)
            logC_series[:, i] = _logmeanexp(sC, axis=1)

    # Convert to linear means per stone across sweeps
    A_series = np.exp(logA_series)  # (nsweeps, m)
    C_series = np.exp(logC_series)
    muA = A_series.mean(axis=0)  # (m,)
    muC = C_series.mean(axis=0)
    
    # Bridge estimate
    z_hat = float(np.sum(np.log(muA) - np.log(muC)))
    
    W_series = np.hstack([A_series, C_series])  # (nsweeps, 2m)
    # grad
    grad = np.concatenate([1.0 / muA, -1.0 / muC])  # for var



    return z_hat, W_series, grad






# actual TI

def get_ti(B, L, pchip=True, ba=None, bb=None, fixed_ladder=True,
           mode='obm', batch_size=None, spe_kernel='bartlett'):
    B, L = order_by_betas(B, L)
    L = L.mean(axis=2)  # shape=(ntemps,nsweeps)

    ntemps, nsweeps = B.shape
    ba, bb = ba, bb
    if ba is None:
        ba = 0
    if bb is None:
        bb = ntemps-1

    if batch_size == None:
        batch_size = int(np.sqrt(ntemps))

    # Get integration beta interval
    if pchip:
        cut_low = B[:, -1][ba]
        cut_high = B[:, -1][bb]

        z_t = _pchip_integral_series_cut(B, L, cut_low, cut_high)  # (nsweeps,)
        B2 = _coarse_rows(B)
        L2 = _coarse_rows(L)
        z2_t = _pchip_integral_series_cut(B2, L2, cut_low, cut_high)  # (nsweeps,)

    else:
        B = B[ba:bb+1, :]
        L = L[ba:bb+1, :]

        z_t = np.trapz(L, B, axis=0)    # (nsweeps,)
        B2 = _coarse_rows_drop_hot(B)
        L2 = _coarse_rows_drop_hot(L)
        z2_t = np.trapz(L2, B2, axis=0)    # (nsweeps,)


    # Evidence estimate
    z_hat = z_t.mean()

    # Discretisation error
    z2_hat  = z2_t.mean()
    err_disc = z2_hat-z_hat

    # Sampling Error
    if mode == 'obm':
        if fixed_ladder:
            Sigma_hat = _obm_univariate(z_t, batch_size=batch_size)  # long-run variance
            err_MC = np.sqrt(Sigma_hat / z_t.size)
        # multivariate
        else:
            S = 0.5 * (L[:-1, :] + L[1:, :])  # (ntemps-1, nsweeps)
            dB = B[1:, :] - B[:-1, :]  # (ntemps-1, nsweeps)
            W_series = (-dB).T  # (nsweeps, ntemps-1)
            S_series = S.T  # (nsweeps, ntemps-1)

            # mOBM on S_series to estimate Sig_S
            Sigma_S = _obm_multivariate(S_series, batch_size=batch_size)

            # Propagate with the *average* per-sweep weights (diminishing adaptation)
            vbar = W_series.mean(axis=0)
            Sigma_hat = float(vbar @ Sigma_S @ vbar)
            err_MC = np.sqrt(Sigma_hat / nsweeps)

    elif mode == 'spe':
        if fixed_ladder:
            Sigma_LR = _spe_univariate(z_t, bandwidth=None, kernel=spe_kernel, prewhiten=False)
            err_MC = np.sqrt(Sigma_LR / z_t.size)
        # multivariate
        else:
            S = 0.5 * (L[:-1, :] + L[1:, :])
            dB = B[1:, :] - B[:-1, :]
            vbar = (-dB).mean(axis=1)
            S_series = S.T

            Sigma_S = _spe_multivariate(S_series, bandwidth=None, kernel=spe_kernel)
            Sigma_z = float(vbar @ Sigma_S @ vbar)
            Sigma_z = max(Sigma_z, 0.0)
            err_MC = float(np.sqrt(Sigma_z / nsweeps))        




    # Combine with discretization error
    err_tot = np.sqrt(err_disc**2 + err_MC**2)
    return z_hat, err_tot


def get_ss(B, L, bridge=True, ba=None, bb=None, fixed_ladder=True,
           mode='obm', batch_size=None, spe_kernel='bartlett'):
    B, L = order_by_betas(B, L)
    ntemps, nsweeps = B.shape
    ba, bb = ba, bb
    if ba is None:
        ba = 0
    if bb is None:
        bb = ntemps

    B = B[ba:bb+1, :]
    L = L[ba:bb+1, :, :]

    if bridge:
        z_hat, W_series, grad = _ss_bridge(B, L, fixed_ladder)
    else:
        z_hat, W_series, grad = _ss_classic(B, L, fixed_ladder)

        
    if mode=='obm':
        Sigma_hat = _obm_multivariate(W_series, batch_size=batch_size)
    elif mode=='spe':
        Sigma_hat = _spe_multivariate(W_series, bandwidth=None, kernel="bartlett")


    cov_bar = Sigma_hat / nsweeps
    var_logZ = float(grad @ cov_bar @ grad)
    err_MC = np.sqrt(max(0.0, var_logZ))

    return z_hat, err_MC



temp_table = np.array([25.2741, 7., 4.47502, 3.5236, 3.0232,
                      2.71225, 2.49879, 2.34226, 2.22198, 2.12628,
                      2.04807, 1.98276, 1.92728, 1.87946, 1.83774,
                      1.80096, 1.76826, 1.73895, 1.7125, 1.68849,
                      1.66657, 1.64647, 1.62795, 1.61083, 1.59494,
                      1.58014, 1.56632, 1.55338, 1.54123, 1.5298,
                      1.51901, 1.50881, 1.49916, 1.49, 1.4813,
                      1.47302, 1.46512, 1.45759, 1.45039, 1.4435,
                      1.4369, 1.43056, 1.42448, 1.41864, 1.41302,
                      1.40761, 1.40239, 1.39736, 1.3925, 1.38781,
                      1.38327, 1.37888, 1.37463, 1.37051, 1.36652,
                      1.36265, 1.35889, 1.35524, 1.3517, 1.34825,
                      1.3449, 1.34164, 1.33847, 1.33538, 1.33236,
                      1.32943, 1.32656, 1.32377, 1.32104, 1.31838,
                      1.31578, 1.31325, 1.31076, 1.30834, 1.30596,
                      1.30364, 1.30137, 1.29915, 1.29697, 1.29484,
                      1.29275, 1.29071, 1.2887, 1.28673, 1.2848,
                      1.28291, 1.28106, 1.27923, 1.27745, 1.27569,
                      1.27397, 1.27227, 1.27061, 1.26898, 1.26737,
                      1.26579, 1.26424, 1.26271, 1.26121,
                      1.25973])



def set_temp_ladder(ntemps, ndims, temp_table=temp_table):
    tstep = temp_table[ndims-1]
    if ndims > len(temp_table):
        tstep = 1.0 + 2.0*np.sqrt(np.log(4.0))/np.sqrt(ndims)
        
    return np.geomspace(1, tstep**(1-ntemps), ntemps)


# @auto-fold regex /^\s*if/ /^\s*else/ /^\s*def/
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# my coding convention
# **EVAL : evaluate the performance of this method
# **RED  : redo this
# **DEB  : debugging needed in this part
# **DEL  : DELETE AT SOME POINT

from copy import deepcopy
import numpy as np


class PTState(object):
    """Parallel Tempering State that manages a list of StatePlus instances."""

    def __init__(self, coords, copy=False):
        if hasattr(coords[0], "coords"):
            self.ntemps = len(coords)
            self.nwalkers, self.ndim = coords[0].coords.shape
            #self.random_state = coords.random_state
        else:
            self.ntemps, self.nwalkers, self.ndim = coords.shape
        
        self.random_state = None
        self.states = [State(coords[t], copy=copy) for t in range(self.ntemps)]


    def __getitem__(self, n):
        return self.states[n]
    
    def __setitem__(self, n, target):
        self.states[n] = target

    def __len__(self):
        return self.ntemps

    def __iter__(self):
        return iter(self.states)


    @property
    def shape(self):
        return self.ntemps, self.nwalkers, self.ndim

    @property
    def betas(self):
        return np.array([state.beta for state in self])

    @betas.setter
    def betas(self, new_betas):
        for t in range(self.ntemps):
            self[t].beta = new_betas[t]
            


class State(object):
    """The state of the ensemble during an MCMC run

    For backwards compatibility, this will unpack into ``coords, log_prob,
    (blobs), random_state`` when iterated over (where ``blobs`` will only be
    included if it exists and is not ``None``).

    Args:
        coords (ndarray[nwalkers, ndim]): The current positions of the walkers
            in the parameter space.
        log_prob (ndarray[nwalkers, ndim], Optional): Log posterior
            probabilities for the  walkers at positions given by ``coords``.
        blobs (Optional): The metadata “blobs” associated with the current
            position. The value is only returned if lnpostfn returns blobs too.
        random_state (Optional): The current state of the random number
            generator.
    """

    __slots__ = "coords", "log_prob", "log_like", "beta", "blobs"

    def __init__(
        self, coords, log_prob=None, log_like=None, blobs=None, beta=1, copy=False
    ):
        dc = deepcopy if copy else lambda x: x

        if hasattr(coords, "coords"):
            self.coords = dc(coords.coords)
            self.log_prob = dc(coords.log_prob)
            self.log_like = dc(coords.log_like)
            self.beta = dc(coords.beta)
            self.blobs = dc(coords.blobs)
            return

        self.coords = dc(np.atleast_2d(coords))
        self.log_prob = dc(log_prob)
        self.log_like = dc(log_like)
        self.blobs = dc(blobs)
        self.beta = dc(beta)

    def __len__(self):
        return 4 if self.blobs is None else 5

    def __repr__(self):
        return "State({0}, log_prob={1}, log_like={2}, beta={3}, blobs={4})".format(
            self.coords, self.log_prob, self.log_like, self.beta, self.blobs
        )

    def __iter__(self):
        if self.blobs is None:
            return iter((self.coords, self.log_prob, self.log_like, self.beta))
        return iter(
            (self.coords, self.log_prob, self.log_like, self.beta, self.blobs)
        )

    def __getitem__(self, index):
        if index < 0:
            return self[len(self) + index]
        if index == 0:
            return self.coords
        elif index == 1:
            return self.log_prob
        elif index == 2:
            return self.log_like
        elif index == 3:
            return self.beta
        elif index == 4 and self.blobs is not None:
            return self.blobs
        raise IndexError("Invalid index '{0}'".format(index))


# @auto-fold regex /^\s*if/ /^\s*else/ /^\s*def/
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# my coding convention
# **EVAL : evaluate the performance of this method
# **RED  : redo this
# **DEB  : debugging needed in this part
# **DEL  : DELETE AT SOME POINT

import numpy as np
from .state import State, PTState

from emcee import autocorr


class PTBackend(object):
    def __init__(self, dtype=None):
        self.initialized = False
        self.random_state = None
        self._list = []
        self.dtype = dtype or np.float64

    def reset(self, ntemps, nwalkers, ndim, smd_hist=False, tsw_hist=False):
        """Initialize the PTBackend and per-temperature backends.

        Args:
            ntemps (int): Number of temperatures.
            nwalkers (int): Number of walkers.
            ndim (int): Number of dimensions.
            smd_hist (bool): Whether to store SMD history.
            tsw_hist (bool): Whether to store temperature swap history.
        """
        self.ntemps = int(ntemps)
        self.nwalkers = int(nwalkers)
        self.ndim = int(ndim)
        self.iteration = 0
        self.ts_accepted = np.zeros(self.ntemps-1, dtype=self.dtype)

        self.tsw_history_bool = tsw_hist
        self.smd_history_bool = smd_hist

        if tsw_hist:
            self.tsw_history = np.empty((0, self.ntemps-1), dtype=self.dtype)
        if smd_hist:
            self.smd_history = np.empty((0, self.ntemps-1), dtype=self.dtype)

        self.backends = [Backend_plus(dtype=self.dtype) for _ in range(self.ntemps)]
        for backend in self.backends:
            backend.reset(self.nwalkers, self.ndim)
        self.initialized = True
        
    def save_step(self, states, ts_accepted, smd=None):
        """Save a step for all temperatures.

        Args:
            states (list of State): States for each temperature.
            accepted (list of ndarray): Acceptance arrays for each temperature.
            ts_accepted (ndarray): Temperature swap acceptance array.
            smd (Optional[ndarray]): SMD history data.
        """

        # Save temperature swapped/adapted state
        for t, backend in enumerate(self):
            backend.chain[-1, :, :] = states[t].coords
            backend.log_like[-1, :] = states[t].log_like
            backend.log_prob[-1, :] = states[t].log_prob

        # Handle PTBackend-specific data
        if self.tsw_history_bool:
            self.tsw_history[self.iteration] = ts_accepted

        if self.smd_history_bool:
            self.smd_history[self.iteration] = smd
        
        self.ts_accepted += ts_accepted
        self.iteration += 1


    def grow(self, ngrow):
        i = ngrow - (len(self.tsw_history) - self.iteration)
        a = np.empty((i, self.ntemps-1), dtype=self.dtype)

        if self.tsw_history_bool:
            self.tsw_history = np.concatenate((self.tsw_history, a), axis=0)
        if self.smd_history_bool:
            self.smd_history = np.concatenate((self.smd_history, a), axis=0)


    def get_autocorr_time(self, discard=0, thin=1, **kwargs):
        return [b.get_autocorr_time(discard=discard, thin=thin, **kwargs) for b in self.backends]


    def get_log_prob(self, **kwargs):
        """Get log probabilities from all temperatures."""
        return self.get_value('log_prob', **kwargs)

    def get_log_like(self, **kwargs):
        """Get log likelihoods from all temperatures."""
        return self.get_value('log_like', **kwargs)

    def get_chain(self, **kwargs):
        """Get chains from all temperatures."""
        return self.get_value('chain', **kwargs)

    def get_betas(self, **kwargs):
        return self.get_value('beta_history', **kwargs)


    def get_value(self, name, **kwargs):
        return np.array([b.get_value(name, **kwargs) for b in self])


    def get_tsw(self, **kwargs):
        # TODO implement this
        return self.get_value_over('tsw_history', **kwargs)
    

    def get_smd(self, **kwargs):
        # TODO implement this
        return self.get_value_over('smd_history', **kwargs)


    def get_value_over(self, name, flat=False, thin=1, discard=0):
        if self.iteration <= 0:
            raise AttributeError(
                "you must run the sampler with "
                "'store == True' before accessing the "
                "results"
            )

        if name == "blobs" and not self.has_blobs():
            return None

        v = getattr(self, name)[discard + thin - 1 : self.iteration : thin]
        if flat:
            s = list(v.shape[1:])
            s[0] = np.prod(v.shape[:2])
            return v.reshape(s)
        return v

    def get_last_sample(self):
        """Access the most recent sample in the chain"""
        if (not self.initialized) or self.iteration <= 0:
            raise AttributeError(
                "you must run the sampler with "
                "'store == True' before accessing the "
                "results"
            )

        states = PTState([b.get_last_sample() for b in self])
        states.random_state = self.random_state
        return states

    @property
    def shape(self):
        """The dimensions of the ensemble `(ntemps, nwalkers, ndim)`."""
        return self.ntemps, self.nwalkers, self.ndim

    @property
    def accepted(self):
        """Acceptance arrays from all temperatures."""
        return np.array([b.accepted for b in self])

        
    def __getitem__(self, n):
        return self.backends[n]


    def __iter__(self):
        return iter(self.backends)



class Backend_plus(object):
    """A backend that extends emcee's Backend with additional storage for log_like and beta_history."""

    def __init__(self, dtype=None):
        self.initialized = False
        if dtype is None:
            dtype = np.float64
        self.dtype = dtype


    def reset(self, nwalkers, ndim):
        """Clear the state of the chain and empty the backend

        Args:
            nwakers (int): The size of the ensemble
            ndim (int): The number of dimensions

        """
        self.nwalkers = int(nwalkers)
        self.ndim = int(ndim)
        self.iteration = 0
        self.accepted = np.zeros(self.nwalkers, dtype=self.dtype)
        self.chain = np.empty((0, self.nwalkers, self.ndim), dtype=self.dtype)
        self.log_like = np.empty((0, self.nwalkers), dtype=self.dtype)
        self.log_prob = np.empty((0, self.nwalkers), dtype=self.dtype)
        self.beta_history = np.empty((0,), dtype=self.dtype)
        self.blobs = None
        self.random_state = None
        self.initialized = True


    def get_log_like(self, **kwargs):
        """Get the chain of log likelihoods evaluated at the MCMC samples

        Args:
            flat (Optional[bool]): Flatten the chain across the ensemble.
                (default: ``False``)
            thin (Optional[int]): Take only every ``thin`` steps from the
                chain. (default: ``1``)
            discard (Optional[int]): Discard the first ``discard`` steps in
                the chain as burn-in. (default: ``0``)

        Returns:
            array[..., nwalkers]: The chain of log likelihoods.
        """
        return self.get_value("log_like", **kwargs)
    
    def get_log_prob(self, **kwargs):
        """Get the chain of log probabilities evaluated at the MCMC samples
        """
        return self.get_value("log_prob", **kwargs)
    
    def get_chain(self, **kwargs):
        """Get the chain of log probabilities evaluated at the MCMC samples
        """
        return self.get_value("chain", **kwargs)

    def get_betas(self, **kwargs):
        """Get the chain of log likelihoods evaluated at the MCMC samples

        Args:
            flat (Optional[bool]): Flatten the chain across the ensemble.
                (default: ``False``)
            thin (Optional[int]): Take only every ``thin`` steps from the
                chain. (default: ``1``)
            discard (Optional[int]): Discard the first ``discard`` steps in
                the chain as burn-in. (default: ``0``)

        Returns:
            array[..., nwalkers]: The chain of log likelihoods.
        """
        return self.get_value("beta_history", **kwargs)

    def get_blobs(self, **kwargs):
        return self.get_value("blobs", **kwargs)

    def get_last_sample(self):
        """Access the most recent sample in the chain"""
        if not self.initialized or self.iteration <= 0:
            raise AttributeError(
                "you must run the sampler with 'store == True' before accessing the results"
            )
        it = self.iteration
        blobs = self.get_blobs(discard=it - 1)
        if blobs is not None:
            blobs = blobs[0]
        return State(
            self.get_chain(discard=it - 1)[0],
            log_prob=self.get_log_prob(discard=it - 1)[0],
            log_like=self.get_log_like(discard=it - 1)[0],
            beta=self.get_betas(discard=it - 1)[0],
            blobs=blobs,
        )


    def get_value(self, name, flat=False, thin=1, discard=0):
        if self.iteration <= 0:
            raise AttributeError(
                "you must run the sampler with "
                "'store == True' before accessing the "
                "results"
            )

        if name == "blobs" and not self.has_blobs():
            return None

        v = getattr(self, name)[discard + thin - 1 : self.iteration : thin]
        if flat:
            s = list(v.shape[1:])
            s[0] = np.prod(v.shape[:2])
            return v.reshape(s)
        return v
    

    def get_autocorr_time(self, discard=0, thin=1, **kwargs):
        """Compute an estimate of the autocorrelation time for each parameter

        Args:
            thin (Optional[int]): Use only every ``thin`` steps from the
                chain. The returned estimate is multiplied by ``thin`` so the
                estimated time is in units of steps, not thinned steps.
                (default: ``1``)
            discard (Optional[int]): Discard the first ``discard`` steps in
                the chain as burn-in. (default: ``0``)

        Other arguments are passed directly to
        :func:`emcee.autocorr.integrated_time`.

        Returns:
            array[ndim]: The integrated autocorrelation time estimate for the
                chain for each parameter.

        """
        x = self.get_chain(discard=discard, thin=thin)
        return thin * autocorr.integrated_time(x, **kwargs)


    @property
    def beta(self):
        """Access the most recent beta in the chain"""
        return self.beta_history[-1]


    @property
    def shape(self):
        """The dimensions of the ensemble ``(nwalkers, ndim)``"""
        return self.nwalkers, self.ndim

    def has_blobs(self):
        """Returns ``True`` if the model includes blobs"""
        return self.blobs is not None

    def _check_blobs(self, blobs):
        has_blobs = self.has_blobs()
        if has_blobs and blobs is None:
            raise ValueError("inconsistent use of blobs")
        if self.iteration > 0 and blobs is not None and not has_blobs:
            raise ValueError("inconsistent use of blobs")

    def grow(self, ngrow, blobs):
        """Expand the storage space by some number of samples

        Args:
            ngrow (int): The number of steps to grow the chain.
            blobs: The current array of blobs. This is used to compute the
                dtype for the blobs array.

        """
        i = ngrow - (len(self.chain) - self.iteration)
        if i > 0:
            self._check_blobs(blobs)
            a = np.empty((i, self.nwalkers, self.ndim), dtype=self.dtype)
            self.chain = np.concatenate((self.chain, a), axis=0)

            a = np.empty((i, self.nwalkers), dtype=self.dtype)
            self.log_prob = np.concatenate((self.log_prob, a), axis=0)
            
            a = np.empty((i, self.nwalkers), dtype=self.dtype)
            self.log_like = np.concatenate((self.log_like, a), axis=0)

            a = np.empty((i), dtype=self.dtype)
            self.beta_history = np.concatenate((self.beta_history, a), axis=0)
        

            if blobs is not None:
                dt = np.dtype((blobs.dtype, blobs.shape[1:]))
                a = np.empty((i, self.nwalkers), dtype=dt)
                if self.blobs is None:
                    self.blobs = a
                else:
                    self.blobs = np.concatenate((self.blobs, a), axis=0)

    def _check(self, state, accepted):
        self._check_blobs(state.blobs)
        nwalkers, ndim = self.shape
        has_blobs = self.has_blobs()
        if state.coords.shape != (nwalkers, ndim):
            raise ValueError(
                "invalid coordinate dimensions; expected {0}".format(
                    (nwalkers, ndim)
                )
            )
        if state.log_prob.shape != (nwalkers,):
            raise ValueError(
                "invalid log probability size; expected {0}".format(nwalkers)
            )
        if state.blobs is not None and not has_blobs:
            raise ValueError("unexpected blobs")
        if state.blobs is None and has_blobs:
            raise ValueError("expected blobs, but none were given")
        if state.blobs is not None and len(state.blobs) != nwalkers:
            raise ValueError(
                "invalid blobs size; expected {0}".format(nwalkers)
            )
        if accepted.shape != (nwalkers,):
            raise ValueError(
                "invalid acceptance size; expected {0}".format(nwalkers)
            )

    def save_step(self, state, accepted):
        """Save a step to the backend

        Args:
            state (State): The :class:`State` of the ensemble.
            accepted (ndarray): An array of boolean flags indicating whether
                or not the proposal for each walker was accepted.

        """
        self._check(state, accepted)

        self.chain[self.iteration, :, :] = state.coords
        self.log_like[self.iteration, :] = state.log_like
        self.log_prob[self.iteration, :] = state.log_prob
        self.beta_history[self.iteration] = state.beta
        if state.blobs is not None:
            self.blobs[self.iteration, :] = state.blobs
        self.accepted += accepted
        self.iteration += 1


    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        pass





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import reddemcee
import math
import time
#import multiprocessing
import emcee
from multiprocessing import Pool

import matplotlib.pyplot as pl


# defining constants
ndim_ = 2
r_ = 2.  # radius
w_ = 0.1  # width
hard_limit = 6
analytic_z = {2:-1.75, 3:-2.84, 5:-5.6736, 10:-14.59, 20:-36.09}

if True:
    limits = [-hard_limit,  hard_limit]
    c1_ = np.zeros(ndim_)
    c1_[0] = -3.5
    c2_ = np.zeros(ndim_)
    c2_[0] = 3.5
    const_ = math.log(1. / math.sqrt(2. * math.pi * w_**2))  # normalization constant

    # log-likelihood of a single shell
    def logcirc(theta, c):
        d = np.sqrt(np.sum((theta - c)**2, axis=-1))  # |theta - c|
        return const_ - (d - r_)**2 / (2. * w_**2)

    # log-likelihood of two shells
    def loglike(theta):
        return np.logaddexp(logcirc(theta, c1_), logcirc(theta, c2_))

    # our prior transform
    def prior_transform(x):
        return (limits[1] - limits[0]) * x - limits[1]

    def reddprior(theta):
        lp = 0.
        for i in range(ndim_):
            if  theta[i] <= limits[0] or limits[1] <= theta[i]:
                return -np.inf
            else:
                lp += 1/(limits[1] - limits[0])
                #lp += 1/(limits[1] - limits[0])
        return np.log(lp)
    
ntemps_ = 12
nwalkers_ = 200
setup = np.array([ntemps_, nwalkers_, 300, 2])

if True:
#with Pool(20) as mypool:
    ntemps, nwalkers, nsweeps, nsteps = setup
    p0 = list(np.random.uniform(limits[0], limits[1], [ntemps, nwalkers, ndim_]))
    my_betas = np.linspace(1, 0, ntemps)

    time_start = time.time()
    #mypool = Pool(ncores_)
    #my_moves = [[(emcee.moves.DEMove(), 0.8),
    #         (emcee.moves.DESnookerMove(), 0.2)] for _ in range(ntemps)]

    sampler = reddemcee.PTSampler(nwalkers, ndim_,
                                  loglike, reddprior, ntemps=ntemps,
                                  #pool=mypool,
                                  adaptative=True,
                                  #moves=my_moves
                                  betas=my_betas,
                                  config_adaptation_decay=1
                                  #betas=np.array([1.00000000e+00, 6.39396482e-01, 3.95339709e-01, 2.51711228e-01,
                                  #                 1.52989587e-01, 9.86033203e-02, 6.04822861e-02, 4.09715831e-02,
                                  #                 2.54359686e-02, 1.49462952e-02, 8.92801860e-03, 5.00517725e-03,
                                  #                 2.90971855e-03, 1.29839577e-03, 1.47444114e-12]),
                                  )#False
    
    sampler.af_history_bool = True

    sampler.config_adaptation_halflife = 500
    sampler.config_adaptation_rate = 60/nwalkers
    sampler.run_mcmc(p0, nsweeps, nsteps)
    
    time_end = time.time()
    tot_time = time_end-time_start
    print(f'Total time: {tot_time} sec')

discard0 = 500

z1 = sampler.thermodynamic_integration_classic(discard=discard0)
z2 = sampler.thermodynamic_integration(discard=discard0)

logls0 = sampler.get_logls(flat=True, discard=discard0)
logls = np.mean(logls0, axis=1)

if True:
    bh = sampler.betas_history
    rh = sampler.ts_ratios_history
    af = sampler.af_history

    fig, axes = pl.subplots(3, 1, figsize=(14, 6), sharex=True)

    bh1 = bh.reshape((setup[2], setup[0]))
    rh1 = rh.reshape((setup[2], setup[0]-1))
    af1 = np.append(np.zeros(setup[0]), af).reshape((setup[2]+1, setup[0]))
    af1 = np.diff(af1, axis=0)

    for i in range(setup[0]-2):
            bh_sel = bh1[:, i]
            b = 1/np.array(bh_sel)
            axes[0].plot(np.arange(setup[2])*setup[3], b)
            axes[0].set_xscale('log')
            axes[0].set_yscale('log')

    for i in np.arange(setup[0]):
        af_sel = af1[:, i]
        axes[1].plot(np.arange(setup[2])*setup[3], af_sel, alpha=0.5)
            
    for i in np.arange(setup[0]-1):
        r = rh1[:, i]
        axes[2].plot(np.arange(setup[2])*setup[3], r, alpha=0.5)
    
    if True:
        axes[2].set_xlabel("N Step")

        axes[0].set_ylabel(r"$\beta^{-1}$")
        axes[1].set_ylabel(r"$\bar{A_{f}}$")
        axes[2].set_ylabel(r"$a_{frac}$")
            
    pl.tight_layout()
    pl.savefig('test_frac.png')

if True:
    my_text = rf'Evidence: {np.round(z2[0], 3)} $\pm$ {np.round(z2[1], 3)}'
    c = ['C0', 'C1', 'C2', 'C4', 'C7', 'C8', 'C9']
    colors = np.array([c,c,c,c,c]).flatten()
    if True:
        fig, ax = pl.subplots()
        for ti in range(ntemps_):
            bet = bh1.T[ti]
            ax.plot(bet, np.ones_like(bet)*logls[ti], colors[ti])
            ax.plot(bet[-1], logls[ti], colors[ti]+'o')

        ylims = ax.get_ylim()
        
        betas0 = [x[-1] for x in bh1.T]

        ax.fill_between(np.append(betas0, 0),
                        np.append(logls, logls[-1]),
                        y2=ylims[1],
                        #color=rc.fg,
                        color='w',
                        alpha=0.25)
        
        ax.set_ylim(ylims)
    if True:
        #ax.scatter([], [], alpha=0, label=my_text)
        pl.legend(loc=4)
        ax.set_xlabel('temps')
        ax.set_ylabel('mean(logls)')
        
        ax.set_xlim([0, 1])
        pl.tight_layout()
        pl.savefig('test_evidence.png',
                   bbox_inches='tight')
        #pl.show()
        pl.close()


#


# @auto-fold regex /^\s*if/ /^\s*else/ /^\s*def/
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# my coding convention
# **EVAL : evaluate the performance of this method
# **RED  : redo this
# **DEB  : debugging needed in this part
# **DEL  : DELETE AT SOME POINT

import numpy as np
from copy import deepcopy

from .state import State

class PTMove(object):
    def __init__(self):
        self.smd_hist = False
        self.D_ = np.array([1])
        self.inverted = False
        self.freeze = False
        
    def adapt(self, state, dSs, copy=True):
        dc = deepcopy if copy else lambda x: x
        
        ntemps, nwalkers, ndim = state.shape
        betas = dc(state.betas)

        deltaTs = np.diff(1 / betas[:-1])
        deltaTs *= np.exp(dSs)
        betas[1:-1] = 1 / (np.cumsum(deltaTs) + 1 / betas[0])
        dbetas = betas - state.betas

        state.betas = betas
        for t in range(ntemps-1, 0, -1):
            state[t].log_prob += state[t].log_like * dbetas[t]

        return state


    def update(self, state, s1, s2):
        pass


    def swap(self, state):
        betas = state.betas
        ntemps, nwalkers, ndim = state.shape
        n_swap_accept = np.zeros(ntemps-1)
        smd = np.zeros(ntemps-1) if self.smd_hist else None
        dbetas = -np.diff(betas)

        for t in self.beta_range(ntemps):
            state_hot = state[t]
            state_cold = state[t-1]

            ll1 = state_hot.log_like  # hot
            ll2 = state_cold.log_like  # cold

            dbeta = dbetas[t-1]  # replace

            raccept = np.log(np.random.uniform(size=nwalkers))
            paccept = dbeta * (ll1 - ll2)

            asel = paccept > raccept
            n_swap_accept[t-1] = np.sum(asel)

            AA = state_hot.coords[asel]
            BB = state_cold.coords[asel]
            if self.smd_hist:
                if asel.any():
                    a_min_b = (AA-BB)/self.D_
                    sup = np.sqrt(np.sum(np.square(a_min_b), axis=1))
                    zeros = np.zeros(int(nwalkers-n_swap_accept[t-1]))
                    smd[t-1] = np.mean(np.append(sup, zeros))
                else:
                    smd[t-1] = 0.

            state_hot.coords[asel], state_cold.coords[asel] = BB, AA
            state_hot.log_prob[asel], state_cold.log_prob[asel] = state_cold.log_prob[asel] - dbeta*ll2[asel], state_hot.log_prob[asel] + dbeta*ll1[asel]
            state_hot.log_like[asel], state_cold.log_like[asel] = ll2[asel], ll1[asel]

        return state, n_swap_accept/nwalkers, smd


    def beta_range(self, ntemps):
        self.inverted = not self.inverted
        return range(1, ntemps) if self.inverted else range(ntemps-1, 0, -1)



class Move(object):
    def tune(self, state, accepted):
        pass

    def update(self, old_state, new_state, accepted, subset=None):
        """Update a given subset of the ensemble with an accepted proposal

        Args:
            coords: The original ensemble coordinates.
            log_probs: The original log probabilities of the walkers.
            blobs: The original blobs.
            new_coords: The proposed coordinates.
            new_log_probs: The proposed log probabilities.
            new_blobs: The proposed blobs.
            accepted: A vector of booleans indicating which walkers were
                accepted.
            subset (Optional): A boolean mask indicating which walkers were
                included in the subset. This can be used, for example, when
                updating only the primary ensemble in a :class:`RedBlueMove`.

        """
        if subset is None:
            subset = np.ones(len(old_state.coords), dtype=bool)

        # integer indexing over boolean masks for better performance?
        subset_indices = np.where(subset)[0]
        accepted_subset = accepted[subset]
        m1 = subset_indices[accepted_subset]
        m2 = np.where(accepted_subset)[0]
        '''
        m1 = subset & accepted
        m2 = accepted[subset]
        '''
        # Update the old_state with the accepted proposals
        old_state.coords[m1] = new_state.coords[m2]
        old_state.log_prob[m1] = new_state.log_prob[m2]
        old_state.log_like[m1] = new_state.log_like[m2]


        if new_state.blobs is not None:
            if old_state.blobs is None:
                raise ValueError(
                    "If you start sampling with a given log_prob, "
                    "you also need to provide the current list of "
                    "blobs at that position."
                )
            old_state.blobs[m1] = new_state.blobs[m2]

        return old_state


class RedBlueMove(Move):
    def __init__(
        self, nsplits=2, randomize_split=True, live_dangerously=False
    ):
        self.nsplits = int(nsplits)
        self.live_dangerously = live_dangerously
        self.randomize_split = randomize_split

    def setup(self, coords):
        pass

    def get_proposal(self, sample, complement, random):
        raise NotImplementedError(
            "The proposal must be implemented by " "subclasses"
        )

    def propose(self, model, state):
        """Generate a proposal for a move and compute the acceptance based on the proposed position.

        Args:
            model (Model): The model containing the log probability function and random number generator.
            state (State): The current state of the walkers, including their coordinates and log probabilities.

        Returns:
            tuple: A tuple containing:
                - state (State): The updated state of the walkers after the proposal.
                - accepted (np.ndarray): A boolean array indicating which proposals were accepted.
        """
        # Check that the dimensions are compatible.
        nwalkers, ndim = state.coords.shape
        if nwalkers < 2 * ndim and not self.live_dangerously:
            raise RuntimeError(
                "It is unadvisable to use a red-blue move "
                "with fewer walkers than twice the number of "
                "dimensions."
            )

        # Run any move-specific setup.
        self.setup(state.coords)

        # Split the ensemble in half and iterate over these two halves.
        accepted = np.zeros(nwalkers, dtype=bool)
        all_inds = np.arange(nwalkers)
        
        inds = all_inds % self.nsplits
        if self.randomize_split:
            model.random.shuffle(inds)

        # Get the two halves of the ensemble.
        #sets = [state.coords[inds == j] for j in range(self.nsplits)]
        for split in range(self.nsplits):
            S1 = inds == split

            s = state.coords[S1]
            c = state.coords[~S1]

            # Get the move-specific proposal.
            q, factors = self.get_proposal(s, c, model.random)

            # Compute the lnprobs of the proposed position.
            new_log_probs, new_log_likes, new_blobs = model.compute_log_prob_fn(q, state.beta)
            #print(f'new_log_probs = {new_log_probs.shape}')

            # Loop over the walkers and update them accordingly.
            # Why this is not vectorised??????
            '''
            for i, (j, f, nlp) in enumerate(
                zip(all_inds[S1], factors, new_log_probs)
            ):
                lnpdiff = f + nlp - state.log_prob[j]
                if lnpdiff > np.log(model.random.rand()):
                    accepted[j] = True
            '''
            lnpdiff = factors + new_log_probs - state.log_prob[S1]
            u = np.log(model.random.rand(len(lnpdiff)))
            accept = lnpdiff > u
            accepted[S1] = accept

            #print(f'subset = {S1.shape}; accepted = {accepted.shape}')
            #print(f'subset = {S1}; accepted = {accepted}')
            #print(f'q = {q.shape}; new_log_probs = {new_log_probs.shape}')
            new_state = State(q, log_prob=new_log_probs, log_like=new_log_likes, beta=state.beta, blobs=new_blobs)
            self.update(state, new_state, accepted, S1)

        return state, accepted


class StretchMove(RedBlueMove):

    def __init__(self, a=2.0, **kwargs):
        self.a = a
        super(StretchMove, self).__init__(**kwargs)

    def get_proposal(self, s, c, random):
        #c = np.concatenate(c, axis=0)
        Ns, Nc = len(s), len(c)
        ndim = s.shape[1]
        zz = ((self.a - 1.0) * random.rand(Ns) + 1) ** 2.0 / self.a
        factors = (ndim - 1.0) * np.log(zz)
        rint = random.randint(Nc, size=(Ns,))
        return c[rint] - (c[rint] - s) * zz[:, None], factors

